"""
A general central radial scanning (or dwelling) instrument class.

"""

import copy
import sys
import warnings

import numpy as np

try:
    import xarray as xr
    _XARRAY_AVAILABLE = True
except ImportError:
    _XARRAY_AVAILABLE = False

from ..config import get_metadata
from ..exceptions import MissingOptionalDependency
from .transforms import antenna_vectors_to_cartesian, cartesian_to_geographic


class RadarSpectra(object):
    """
    A class for storing antenna coordinate radar spectra data.
    The structure of the Radar class is based on the CF/Radial Data file
    format. Global attributes and variables (section 4.1 and 4.3) are
    represented as a dictionary in the metadata attribute. Other required and
    optional variables are represented as dictionaries in a attribute with the
    same name as the variable in the CF/Radial standard. When a optional
    attribute not present the attribute has a value of None. The data for a
    given variable is stored in the dictionary under the 'data' key. Moment
    field data is stored as a dictionary of dictionaries in the fields
    attribute. Sub-convention variables are stored as a dictionary of
    dictionaries under the meta_group attribute.

    Refer to the attribute section for information on the parameters.

    Attributes
    ----------
    time : dict
        Time at the center of each ray.
    range : dict
        Range to the center of each gate (bin).
    fields : dict of dicts
        Spectra fields.
    metadata : dict
        Metadata describing the instrument and data.
    scan_type : str
        Type of scan, one of 'ppi', 'rhi', 'sector' or 'other'. If the scan
        volume contains multiple sweep modes this should be 'other'.
    latitude : dict
        Latitude of the instrument.
    longitude : dict
        Longitude of the instrument.
    altitude : dict
        Altitude of the instrument, above sea level.
    altitude_agl : dict or None
        Altitude of the instrument above ground level. If not provided this
        attribute is set to None, indicating this parameter not available.
    sweep_number : dict
        The number of the sweep in the volume scan, 0-based.
    sweep_mode : dict
        Sweep mode for each mode in the volume scan.
    fixed_angle : dict
        Target angle for thr sweep. Azimuth angle in RHI modes, elevation
        angle in all other modes.
    sweep_start_ray_index : dict
        Index of the first ray in each sweep relative to the start of the
        volume, 0-based.
    sweep_end_ray_index : dict
        Index of the last ray in each sweep relative to the start of the
        volume, 0-based.
    rays_per_sweep : LazyLoadDict
        Number of rays in each sweep. The data key of this attribute is
        create upon first access from the data in the sweep_start_ray_index and
        sweep_end_ray_index attributes. If the sweep locations needs to be
        modified, do this prior to accessing this attribute or use
        :py:func:`init_rays_per_sweep` to reset the attribute.
    target_scan_rate : dict or None
        Intended scan rate for each sweep. If not provided this attribute is
        set to None, indicating this parameter is not available.
    rays_are_indexed : dict or None
        Indication of whether ray angles are indexed to a regular grid in
        each sweep. If not provided this attribute is set to None, indicating
        ray angle spacing is not determined.
    ray_angle_res : dict or None
        If rays_are_indexed is not None, this provides the angular resolution
        of the grid. If not provided or available this attribute is set to
        None.
    azimuth : dict
        Azimuth of antenna, relative to true North. Azimuth angles are
        recommended to be expressed in the range of [0, 360], but other
        representations are not forbidden.
    elevation : dict
        Elevation of antenna, relative to the horizontal plane. Elevation
        angles are recommended to be expressed in the range of [-180, 180],
        but other representations are not forbidden.
    gate_x, gate_y, gate_z : LazyLoadDict
        Location of each gate in a Cartesian coordinate system assuming a
        standard atmosphere with a 4/3 Earth's radius model. The data keys of
        these attributes are create upon first access from the data in the
        range, azimuth and elevation attributes. If these attributes are
        changed use :py:func:`init_gate_x_y_z` to reset.
    gate_longitude, gate_latitude : LazyLoadDict
        Geographic location of each gate. The projection parameter(s) defined
        in the `projection` attribute are used to perform an inverse map
        projection from the Cartesian gate locations relative to the radar
        location to longitudes and latitudes. If these attributes are changed
        use :py:func:`init_gate_longitude_latitude` to reset the attributes.
    projection : dic or str
        Projection parameters defining the map projection used to transform
        from Cartesian to geographic coordinates. The default dictionary sets
        the 'proj' key to 'pyart_aeqd' indicating that the native Py-ART
        azimuthal equidistant projection is used. This can be modified to
        specify a valid pyproj.Proj projparams dictionary or string.
        The special key '_include_lon_0_lat_0' is removed when interpreting
        this dictionary. If this key is present and set to True, which is
        required when proj='pyart_aeqd', then the radar longitude and
        latitude will be added to the dictionary as 'lon_0' and 'lat_0'.
    gate_altitude : LazyLoadDict
        The altitude of each radar gate as calculated from the altitude of the
        radar and the Cartesian z location of each gate. If this attribute
        is changed use :py:func:`init_gate_altitude` to reset the attribute.
    scan_rate : dict or None
        Actual antenna scan rate. If not provided this attribute is set to
        None, indicating this parameter is not available.
    antenna_transition : dict or None
        Flag indicating if the antenna is in transition, 1 = yes, 0 = no.
        If not provided this attribute is set to None, indicating this
        parameter is not available.
    rotation : dict or None
        The rotation angle of the antenna. The angle about the aircraft
        longitudinal axis for a vertically scanning radar.
    tilt : dict or None
        The tilt angle with respect to the plane orthogonal (Z-axis) to
        aircraft longitudinal axis.
    roll : dict or None
        The roll angle of platform, for aircraft right wing down is positive.
    drift : dict or None
        Drift angle of antenna, the angle between heading and track.
    heading : dict or None
        Heading (compass) angle, clockwise from north.
    pitch : dict or None
        Pitch angle of antenna, for aircraft nose up is positive.
    georefs_applied : dict or None
        Indicates whether the variables have had georeference calculation
        applied.  Leading to Earth-centric azimuth and elevation angles.
    instrument_parameters : dict of dicts or None
        Instrument parameters, if not provided this attribute is set to None,
        indicating these parameters are not avaiable. This dictionary also
        includes variables in the radar_parameters CF/Radial subconvention.
    radar_calibration : dict of dicts or None
        Instrument calibration parameters. If not provided this attribute is
        set to None, indicating these parameters are not available
    ngates : int
        Number of gates (bins) in a ray.
    nrays : int
        Number of rays in the volume.
    nsweeps : int
        Number of sweep in the volume.

    """
    def __init__(self, time, _range, fields, metadata, scan_type,
                 latitude, longitude, altitude,
                 sweep_number, sweep_mode, fixed_angle, sweep_start_ray_index,
                 sweep_end_ray_index,
                 azimuth, elevation, npulses_max, velocity_bins,
                 altitude_agl=None,
                 target_scan_rate=None, rays_are_indexed=None,
                 ray_angle_res=None,
                 scan_rate=None, antenna_transition=None,
                 instrument_parameters=None,
                 radar_calibration=None, georefs_applied=None
                ):
        warnings.warn("Radar Spectra object is in early development, "
                      "errors may arise, use at your own risk! ")
        if not _XARRAY_AVAILABLE:
            raise MissingOptionalDependency(
                "Xarray is required to use RadarSpectra but is "
                "not installed!")
        self.field_names = ['spectra']

        self.ds = xr.Dataset(
            data_vars={
                'spectra': (('time', 'range', 'npulses_max'), fields),
                'velocity_bins': velocity_bins,
                'scan_type': scan_type,
                'latitude': latitude,
                'longitude': longitude,
                'altitude': altitude,
                'sweep_number': sweep_number,
                'sweep_mode': sweep_mode,
                'fixed_angle': fixed_angle,
                'sweep_start_ray_index': sweep_start_ray_index,
                'sweep_end_ray_index': sweep_end_ray_index,
                'azimuth': azimuth,
                'elevation': elevation},
            coords={'time': time,
                    'range': _range,
                    'npulses_max': npulses_max},
            attrs=metadata)

        self.ds['ngates'] = len(_range.values)
        self.ds['nrays'] = len(time.values)
        self.ds['nsweeps'] = len(sweep_number.values)
        self.ds.attrs['projection'] = {'proj': 'pyart_aeqd',
                                       '_include_lon_0_lat_0': True}

        # initalize attributes with lazy load dictionaries
        self.init_rays_per_sweep()
        self.init_gate_x_y_z()
        self.init_gate_longitude_latitude()
        self.init_gate_altitude()

    @property
    def fields(self):
        field_dict = {}
        for key in self.field_names:
            if key in self.ds.variables.keys():
                field_dict[key] = self.ds[key]
        return xr.Dataset(field_dict)

    @property
    def time(self):
        return self.ds.time

    @property
    def range(self):
        return self.ds.range

    @property
    def npulses_max(self):
        return self.ds.npulses_max

    @property
    def velocity_bins(self):
        return self.ds.velocity_bins

    @property
    def latitude(self):
        return self.ds.latitude

    @property
    def longitude(self):
        return self.ds.longitude

    @property
    def altitude(self):
        return self.ds.altitude

    @property
    def fixed_angle(self):
        return self.ds.fixed_angle

    @property
    def sweep_mode(self):
        return self.ds.sweep_mode

    @property
    def sweep_number(self):
        return self.ds.sweep_number

    @property
    def scan_type(self):
        return self.ds.scan_type

    @property
    def elevation(self):
        return self.ds.elevation

    @property
    def azimuth(self):
        return self.ds.azimuth

    @property
    def sweep_start_ray_index(self):
        return self.ds.sweep_start_ray_index

    @property
    def sweep_end_ray_index(self):
        return self.ds.sweep_end_ray_index

    @property
    def rays_per_sweep(self):
        return self.ds.rays_per_sweep

    @property
    def gate_x(self):
        return self.ds.gate_x

    @property
    def gate_y(self):
        return self.ds.gate_y

    @property
    def gate_z(self):
        return self.ds.gate_z

    @property
    def gate_latitude(self):
        return self.ds.gate_latitude

    @property
    def gate_longitude(self):
        return self.ds.gate_longitude

    @property
    def gate_altitude(self):
        return self.ds.gate_altitude

    @property
    def ngates(self):
        return self.ds.ngates

    @property
    def nrays(self):
        return self.ds.nrays

    @property
    def nsweeps(self):
        return self.ds.nsweeps

    @property
    def projection(self):
        return self.ds.attrs['projection']

    def init_rays_per_sweep(self):
        """ Initialize or reset the rays_per_sweep attribute. """
        _rays_per_sweep_data_factory(self.ds)

    def init_gate_x_y_z(self):
        """ Initialize or reset the gate_{x, y, z} attributes. """
        _gate_data_factory(self.ds)

    def init_gate_longitude_latitude(self):
        """
        Initialize or reset the gate_longitude and gate_latitude attributes.
        """
        _gate_lon_lat_data_factory(self.ds)

    def init_gate_altitude(self):
        """ Initialize the gate_altitude attribute. """
        _gate_altitude_data_factory(self.ds)

    def _check_sweep_in_range(self, sweep):
        """ Check that a sweep number is in range. """
        if sweep < 0 or sweep >= self.nsweeps:
            raise IndexError('Sweep out of range: ', sweep)

        # get methods

    def get_start(self, sweep):
        """ Return the starting ray index for a given sweep. """
        self._check_sweep_in_range(sweep)
        return self.sweep_start_ray_index.values[sweep]

    def get_end(self, sweep):
        """ Return the ending ray for a given sweep. """
        self._check_sweep_in_range(sweep)
        return self.sweep_end_ray_index.values[sweep]

    def get_start_end(self, sweep):
        """ Return the starting and ending ray for a given sweep. """
        return self.get_start(sweep), self.get_end(sweep)

    def get_slice(self, sweep):
        """ Return a slice for selecting rays for a given sweep. """
        start, end = self.get_start_end(sweep)
        return slice(start, end+1)

    def check_field_exists(self, field_name):
        """
        Check that a field exists in the fields dictionary.
        If the field does not exist raise a KeyError.

        Parameters
        ----------
        field_name : str
            Name of field to check.

        """
        if field_name not in self.fields.keys():
            raise KeyError('Field not available: ' + field_name)

    # Iterators
    def iter_start(self):
        """ Return an iterator over the sweep start indices. """
        return (s for s in self.sweep_start_ray_index.values)

    def iter_end(self):
        """ Return an iterator over the sweep end indices. """
        return (s for s in self.sweep_end_ray_index.values)

    def iter_start_end(self):
        """ Return an iterator over the sweep start and end indices. """
        return ((s, e) for s, e in zip(self.iter_start(), self.iter_end()))

    def iter_slice(self):
        """ Return an iterator which returns sweep slice objects. """
        return (slice(s, e+1) for s, e in self.iter_start_end())

    def iter_field(self, field_name):
        """ Return an iterator which returns sweep field data. """
        self.check_field_exists(field_name)
        return (self.fields[field_name].values[s] for s in self.iter_slice())

    def iter_azimuth(self):
        """ Return an iterator which returns sweep azimuth data. """
        return (self.azimuth.values[s] for s in self.iter_slice())

    def iter_elevation(self):
        """ Return an iterator which returns sweep elevation data. """
        return (self.elevation.values[s] for s in self.iter_slice())

    def to_vpt(self):
        """ Returns a simple Radar object in VPT scan type with spectra moments
        such as reflectivity and mean velocity. """
        from ..testing import make_empty_ppi_radar
        from ..retrieve import spectra_moments
        from ..util import to_vpt

        rng_len = len(self.range.values)
        time_len = len(self.time.values)
        vpt_radar = make_empty_ppi_radar(
            ngates=rng_len, rays_per_sweep=time_len, nsweeps=1)

        fields = spectra_moments(self)

        rng_dict = get_metadata('range')
        rng_dict['data'] = self.range.values

        time_dict = get_metadata('time')
        time_dict['data'] = self.time.values

        vpt_radar.range = rng_dict
        vpt_radar.time = time_dict
        vpt_radar.fields = fields
        vpt_radar.metadata['instrument_name'] = 'KAZR'
        to_vpt(vpt_radar)
        return vpt_radar


def _rays_per_sweep_data_factory(radar):
    """ Return a function which returns the number of rays per sweep. """
    rays_per_sweep_dict = get_metadata('rays_per_sweep')
    rays_per_sweep = (
        radar.sweep_end_ray_index.values -
        radar.sweep_start_ray_index.values + 1)
    radar['rays_per_sweep'] = xr.DataArray(np.array(rays_per_sweep),
                                           attrs=rays_per_sweep_dict)


def _gate_data_factory(radar):
    """ Return a function which returns the Cartesian locations of gates. """
    ranges = radar.range.values
    azimuths = radar.azimuth.values
    elevations = radar.elevation.values
    cartesian_coords = antenna_vectors_to_cartesian(
        ranges, azimuths, elevations, edges=False)

    # load x, y, and z data except for the coordinate in question
    gate_x_dict = get_metadata('gate_x')
    radar['gate_x'] = xr.DataArray(cartesian_coords[0],
                                   dims=('time', 'range'),
                                   attrs=gate_x_dict)
    gate_y_dict = get_metadata('gate_y')
    radar['gate_y'] = xr.DataArray(cartesian_coords[1],
                                   dims=('time', 'range'),
                                   attrs=gate_y_dict)
    gate_z_dict = get_metadata('gate_z')
    radar['gate_z'] = xr.DataArray(cartesian_coords[2],
                                   dims=('time', 'range'),
                                   attrs=gate_z_dict)

def _gate_lon_lat_data_factory(radar):
    """ Return a function which returns the geographic locations of gates. """
    x = radar.gate_x.values
    y = radar.gate_y.values
    projparams = radar.projection.copy()
    if projparams.pop('_include_lon_0_lat_0', False):
        projparams['lon_0'] = radar.longitude.values
        projparams['lat_0'] = radar.latitude.values
    geographic_coords = cartesian_to_geographic(x, y, projparams)
    # set the other geographic coordinate
    gate_latitude_dict = get_metadata('gate_latitude')
    radar['gate_latitude'] = xr.DataArray(geographic_coords[1],
                                          dims=('time', 'range'),
                                          attrs=gate_latitude_dict)
    gate_longitude_dict = get_metadata('gate_longitude')
    radar['gate_longitude'] = xr.DataArray(geographic_coords[0],
                                           dims=('time', 'range'),
                                           attrs=gate_longitude_dict)

def _gate_altitude_data_factory(radar):
    """ Return a function which returns the gate altitudes. """
    try:
        alt = radar.altitude.values + radar.gate_z.values
    except ValueError:
        alt = np.mean(radar.altitude.values) + radar.gate_z.values
    gate_altitude_dict = get_metadata('gate_altitude')
    radar['gate_altitude'] = xr.DataArray(alt, dims=('time', 'range'),
                                          attrs=gate_altitude_dict)



""" Unit Tests for Py-ART's io/cfradial.py module. """

import warnings

import numpy as np
from numpy.ma.core import MaskedArray
from numpy.testing import assert_array_equal, assert_almost_equal
import netCDF4
import pytest

import pyart

#################################################
# read_cfradial tests (verify radar attributes) #
#################################################

# read in the sample file and create a a Radar object
radar = pyart.io.read_cfradial(pyart.testing.CFRADIAL_PPI_FILE)


# time attribute
def test_time():
    assert 'comment' in radar.time.keys()
    assert 'long_name' in radar.time.keys()
    assert 'standard_name' in radar.time.keys()
    assert 'units' in radar.time.keys()
    assert 'calendar' in radar.time.keys()
    assert 'data' in radar.time.keys()
    assert radar.time['units'] == 'seconds since 2011-05-20T10:54:08Z'
    assert radar.time['data'].shape == (40, )
    assert_almost_equal(radar.time['data'][38], 7, 0)


# range attribute
def test_range():
    assert 'long_name' in radar.range
    assert 'standard_name' in radar.range
    assert 'meters_to_center_of_first_gate' in radar.range
    assert 'meters_between_gates' in radar.range
    assert 'units' in radar.range
    assert 'data' in radar.range
    assert 'spacing_is_constant' in radar.range
    assert radar.range['data'].shape == (42, )
    assert_almost_equal(radar.range['data'][0], 0, 0)


# fields attribute is tested later


# metadata attribute
def test_metadata():
    assert 'Conventions' in radar.metadata
    assert 'comment' in radar.metadata
    assert 'history' in radar.metadata
    assert 'institution' in radar.metadata
    assert 'instrument_name' in radar.metadata
    assert 'instrument_type' in radar.metadata
    assert 'platform_type' in radar.metadata
    assert 'primary_axis' in radar.metadata
    assert 'references' in radar.metadata
    assert 'source' in radar.metadata
    assert 'title' in radar.metadata
    assert 'version' in radar.metadata
    assert 'volume_number' in radar.metadata


# scan_type attribute
def test_scan_type():
    assert radar.scan_type == 'ppi'


# latitude attribute
def test_latitude():
    assert 'data' in radar.latitude
    assert 'standard_name' in radar.latitude
    assert 'units' in radar.latitude
    assert radar.latitude['data'].shape == (1, )
    assert_almost_equal(radar.latitude['data'], 36.0, 0)


# longitude attribute
def test_longitude():
    assert 'data' in radar.longitude
    assert 'standard_name' in radar.longitude
    assert 'units' in radar.longitude
    assert radar.longitude['data'].shape == (1, )
    assert_almost_equal(radar.longitude['data'], -98, 0)


# altitude attribute
def test_altitude():
    assert 'data' in radar.altitude
    assert 'standard_name' in radar.altitude
    assert 'units' in radar.altitude
    assert 'positive' in radar.altitude
    assert radar.altitude['data'].shape == (1, )
    assert_almost_equal(radar.altitude['data'], 214, 0)


# altitude_agl attribute
def test_altitude_agl():
    assert radar.altitude_agl is None


# sweep_number attribute
def test_sweep_number():
    assert 'standard_name' in radar.sweep_number
    assert np.all(radar.sweep_number['data'] == range(1))


# sweep_mode attribute
def test_sweep_mode():
    assert 'standard_name' in radar.sweep_mode
    assert radar.sweep_mode['data'].shape == (1, 32)
    assert radar.sweep_mode['data'].dtype.char == 'S'
    str_array = netCDF4.chartostring(radar.sweep_mode['data'])
    try:
        assert np.all(str_array == ['azimuth_surveillance'])
    except AssertionError:
        assert np.all(str_array == [b'azimuth_surveillance'])


# fixed_angle attribute
def test_fixed_angle():
    assert 'standard_name' in radar.fixed_angle
    assert 'units' in radar.fixed_angle
    assert radar.fixed_angle['data'].shape == (1, )
    assert_almost_equal(radar.fixed_angle['data'][0], 0.50, 2)


# sweep_start_ray_index attribute
def test_sweep_start_ray_index():
    assert 'long_name' in radar.sweep_start_ray_index
    assert radar.sweep_start_ray_index['data'].shape == (1, )
    assert_almost_equal(radar.sweep_start_ray_index['data'][0], 0, 0)


# sweep_end_ray_index attribute
def test_sweep_end_ray_index():
    assert 'long_name' in radar.sweep_end_ray_index
    assert radar.sweep_end_ray_index['data'].shape == (1, )
    assert_almost_equal(radar.sweep_end_ray_index['data'][0], 39, 0)


# target_scan_rate attribute
def test_target_scan_rate():
    assert radar.target_scan_rate is None


# azimuth attribute
def test_azimuth():
    assert 'standard_name' in radar.azimuth
    assert 'units' in radar.azimuth
    assert_almost_equal(radar.azimuth['data'][0], 360, 0)
    assert_almost_equal(radar.azimuth['data'][10], 90., 0)


# elevation attribute
def test_elevation():
    assert 'standard_name' in radar.elevation
    assert 'units' in radar.elevation
    assert radar.elevation['data'].shape == (40, )
    assert_almost_equal(radar.elevation['data'][0], 0.48, 2)


# scan_rate attribute
def test_scan_rate():
    assert radar.scan_rate is None


# antenna_transition attribute
def test_antenna_transition():
    assert radar.antenna_transition is None


# instrument_parameters attribute
@pytest.mark.parametrize(
    "keys", ['prt_mode', 'prt', 'nyquist_velocity', 'unambiguous_range'])
def test_instument_parameters(keys):
    # instrument_parameter sub-convention
    description = 'instrument_parameters: %s' % keys
    check_instrument_parameter.description = description
    check_instrument_parameter(keys)


def check_instrument_parameter(param):
    assert param in radar.instrument_parameters
    param_dic = radar.instrument_parameters[param]
    assert param_dic['meta_group'] == 'instrument_parameters'


# ngates attribute
def test_ngates():
    assert radar.ngates == 42


# nrays attribute
def test_nrays():
    assert radar.nrays == 40


# nsweeps attribute
def test_nsweeps():
    assert radar.nsweeps == 1


####################
# fields attribute #
####################


@pytest.mark.parametrize("field", ['reflectivity_horizontal', ])
def test_field_dics(field):
    description = "field : %s, dictionary" % field
    check_field_dic.description = description
    check_field_dic(field)


def check_field_dic(field):
    """ Check that the required keys are present in a field dictionary. """
    assert 'standard_name' in radar.fields[field]
    assert 'units' in radar.fields[field]
    assert '_FillValue' in radar.fields[field]
    assert 'coordinates' in radar.fields[field]


@pytest.mark.parametrize("field", ['reflectivity_horizontal', ])
def test_field_shapes(field):
    description = "field : %s, shape" % field
    check_field_shape.description = description
    check_field_shape(field)


def check_field_shape(field):
    assert radar.fields[field]['data'].shape == (40, 42)


fields = {'reflectivity_horizontal': MaskedArray, }
@pytest.mark.parametrize(
    "field, field_type", fields.items(), ids=list(fields.keys()))
def test_field_types(field, field_type):
    description = "field : %s, type" % field
    check_field_type.description = description
    check_field_type(field, field_type)


def check_field_type(field, field_type):
    assert type(radar.fields[field]['data']) is field_type


fields = {'reflectivity_horizontal': -6.0, }
@pytest.mark.parametrize(
    "field, field_value", fields.items(), ids=list(fields.keys()))
def test_field_first_points(field, field_value):
    # these values can be found using:
    # [round(radar.fields[f]['data'][0,0]) for f in radar.fields]
    description = "field : %s, first point" % field
    check_field_first_point.description = description
    check_field_first_point(field, field_value)


def check_field_first_point(field, value):
    assert_almost_equal(radar.fields[field]['data'][0, 0], value, 0)

########################################################################
# write_cfradial tests (verify data in written netCDF matches original #
########################################################################


def test_write_ppi():
    # CF/Radial example file -> Radar object -> netCDF file
    with pyart.testing.InTemporaryDirectory():
        tmpfile = 'tmp_ppi.nc'
        radar = pyart.io.read_cfradial(pyart.testing.CFRADIAL_PPI_FILE)
        pyart.io.write_cfradial(tmpfile, radar)
        ref = netCDF4.Dataset(pyart.testing.CFRADIAL_PPI_FILE)
        dset = netCDF4.Dataset(tmpfile)
        check_dataset_to_ref(dset, ref)
        dset.close()


def test_write_ppi_unicode():
    # GitHub issue #381, unicode dtypes cause write to fail
    # CF/Radial example file -> Radar object -> netCDF file
    with pyart.testing.InTemporaryDirectory():
        tmpfile = 'tmp_ppi_unicode.nc'
        radar = pyart.io.read_cfradial(pyart.testing.CFRADIAL_PPI_FILE)
        # force the sweep mode array to be of unicode
        radar.sweep_mode['data'] = np.array(
            ['azimuth_surveillance'], dtype='U')
        pyart.io.write_cfradial(tmpfile, radar)
        ref = netCDF4.Dataset(pyart.testing.CFRADIAL_PPI_FILE)
        dset = netCDF4.Dataset(tmpfile)
        check_dataset_to_ref(dset, ref)
        dset.close()


def test_write_ppi_U1():
    # GitHub issue #381, unicode dtypes cause write to fail
    # CF/Radial example file -> Radar object -> netCDF file
    with pyart.testing.InTemporaryDirectory():
        tmpfile = 'tmp_ppi_unicode.nc'
        radar = pyart.io.read_cfradial(pyart.testing.CFRADIAL_PPI_FILE)
        # force the sweep mode array to be a masked U1 array
        radar.sweep_mode['data'] = radar.sweep_mode['data'].astype('U')
        pyart.io.write_cfradial(tmpfile, radar)
        ref = netCDF4.Dataset(pyart.testing.CFRADIAL_PPI_FILE)
        dset = netCDF4.Dataset(tmpfile)
        check_dataset_to_ref(dset, ref)
        dset.close()


def test_write_ppi_unknown_instrument_parameter_element():
    # CF/Radial example file -> Radar object -> netCDF file
    with pyart.testing.InTemporaryDirectory():
        tmpfile = 'tmp_ppi_unknonw_ip.nc'
        radar = pyart.io.read_cfradial(pyart.testing.CFRADIAL_PPI_FILE)
        radar.instrument_parameters['foobar'] = {'data': np.zeros(40)}
        pytest.warns(UserWarning, pyart.io.write_cfradial, tmpfile, radar)
        ref = netCDF4.Dataset(pyart.testing.CFRADIAL_PPI_FILE)
        dset = netCDF4.Dataset(tmpfile)
        check_dataset_to_ref(dset, ref)
        dset.close()


def test_write_rhi():
    # CF/Radial example file -> Radar object -> netCDF file
    with pyart.testing.InTemporaryDirectory():
        tmpfile = 'tmp_rhi.nc'
        radar = pyart.io.read_cfradial(pyart.testing.CFRADIAL_RHI_FILE)
        pyart.io.write_cfradial(tmpfile, radar)
        ref = netCDF4.Dataset(pyart.testing.CFRADIAL_RHI_FILE)
        dset = netCDF4.Dataset(tmpfile)
        check_dataset_to_ref(dset, ref)
        dset.close()


def test_write_ppi_arm_time_vars():
    # CF/Radial example file -> Radar object -> netCDF file
    with pyart.testing.InTemporaryDirectory():
        tmpfile = 'tmp_arm_time_vars.nc'
        radar = pyart.io.read_cfradial(pyart.testing.CFRADIAL_PPI_FILE)
        pyart.io.write_cfradial(tmpfile, radar, arm_time_variables=True)
        dset = netCDF4.Dataset(tmpfile)
        assert 'base_time' in dset.variables
        assert 'time_offset' in dset.variables

        base_time = dset.variables['base_time']
        assert base_time[:] == 1305888856
        # The month is abbreviated according to the current locale, for
        # en_US the complete string is 20-May-2011,10:54:16 GMT.
        # In order to support other locales test only the non-locale dependent
        # portions of the string.
        assert base_time.string.startswith('20-')
        assert base_time.string.endswith('2011,10:54:16 GMT')

        time_offset = dset.variables['time_offset']
        assert time_offset.units == 'seconds since 2011-05-20 10:54:08'
        assert_almost_equal(time_offset[10], 11, 0)
        dset.close()


def check_dataset_to_ref(dset, ref):
    """ Check that all data in Dataset is contained in the ref Dataset. """

    # file format (these are not expected to match)
    assert dset.file_format == ref.file_format

    # global attributes
    dset_attributes = dset.ncattrs()
    ref_attributes = ref.ncattrs()

    for attr in dset_attributes:
        print("Global attribute:", attr)
        assert attr in ref_attributes
        attribute_equal(dset, ref, attr, allow_str_case_diff=True)

    # cmptypes, expected to be empty
    print("Checking cmptypes...")
    assert dset.cmptypes == ref.cmptypes

    # groups, expected to be empty
    print("Checking groups...")
    assert dset.groups == ref.groups

    # dimensions (these are not expected to match)
    print("Checking dimensions")
    for dim in dset.dimensions.keys():
        print("Dimension", dim)
        assert dim in ref.dimensions.keys()

    # variables
    print("Checking variables...")
    dset_vars = dset.variables
    ref_vars = ref.variables
    for v in dset_vars.keys():
        if v in ['platform_type', 'instrument_type', 'primary_axis']:
            continue    # default value created by pyart.
        print("Variable", v)
        check_variable_to_ref(dset_vars[v], ref_vars[v])


def check_variable_to_ref(var, ref_var):
    """ Check that the data/metadata in var matches the ref. variable. """
    # check variable attributes
    for attr in var.ncattrs():
        print("Checking attribute", attr)
        assert attr in ref_var.ncattrs()
        attribute_equal(var, ref_var, attr)

    # instance variables
    assert var.dimensions == ref_var.dimensions
    assert var.dtype == ref_var.dtype
    assert var.ndim == ref_var.ndim
    assert var.shape == ref_var.shape

    if 'least_significant_digit' in dir(var):
        assert 'least_significant_digit' in dir(ref_var)
        assert var.least_significant_digit == ref_var.least_significant_digit

    # properties
    assert var.size == ref_var.size
    # netCDF4 version < 1.1.1 use the maskandscale attribute
    if hasattr(ref_var, 'maskandscale'):
        assert var.maskandscale == ref_var.maskandscale
    # netCDF4 version >= 1.1.1 use a seperate mask and scale attributes
    if hasattr(ref_var, 'mask'):
        assert var.mask == ref_var.mask
    if hasattr(ref_var, 'scale'):
        assert var.scale == ref_var.scale

    # data and the mask
    data = var[:]
    ref_data = ref_var[:]
    assert ('mask' in dir(data)) == ('mask' in dir(ref_data))
    if 'mask' in dir(data):
        assert_array_equal(data, ref_data)
        assert_array_equal(data.mask, ref_data.mask)
    else:
        assert_array_equal(data, ref_data)


def attribute_equal(class1, class2, key, allow_str_case_diff=True):
    """ Check that an attribute of two classes are equal. """
    a1 = getattr(class1, key)
    a2 = getattr(class2, key)

    assert type(a1) == type(a2)

    if type(a1) is str and allow_str_case_diff:
        assert a1.upper() == a2.upper()
    else:
        assert a1 == a2
    return


def test_auto_history_and_conventions():
    # history and Conventions metadata should be created on write if
    # they do not exist in the original radar object
    with pyart.testing.InTemporaryDirectory():
        tmpfile = 'tmp_auto_history.nc'
        radar = pyart.io.read_cfradial(pyart.testing.CFRADIAL_RHI_FILE)
        radar.metadata.pop('Conventions')
        radar.metadata.pop('history')
        pyart.io.write_cfradial(tmpfile, radar)
        radar2 = pyart.io.read_cfradial(tmpfile)
        assert 'Conventions' in radar2.metadata
        assert 'history' in radar2.metadata


def test_rhi_sweep_end_ray_index():
    radar = pyart.io.read_cfradial(pyart.testing.CFRADIAL_RHI_FILE)
    assert 'long_name' in radar.sweep_end_ray_index
    assert radar.sweep_end_ray_index['data'].shape == (1, )
    assert_almost_equal(radar.sweep_end_ray_index['data'][0], 39, 0)


def test_delay_field_loading():
    radar = pyart.io.read_cfradial(
        pyart.testing.CFRADIAL_PPI_FILE, delay_field_loading=True)
    assert isinstance(radar.fields['reflectivity_horizontal'],
                      pyart.lazydict.LazyLoadDict)
    data = radar.fields['reflectivity_horizontal']['data']
    assert isinstance(data, MaskedArray)
    assert data.shape == (40, 42)
    assert_almost_equal(data[0, 0], -6.0, 0)


def test_create_ncvar_different_dtype():
    # test _Write_as_dtype key handling in _create_ncvar
    with pyart.testing.InTemporaryDirectory():
        tmpfile = 'tmp_nvcar_different_dtype.nc'
        dset = netCDF4.Dataset(tmpfile, mode='w')
        dset.createDimension('x', 256)
        dic = {
            'data': np.arange(256, dtype='float32') * 0.5 + 100.,
            '_Write_as_dtype': 'u1',
            '_FillValue': 100.,
        }
        assert 'add_offset' not in dic
        assert 'scale_factor' not in dic

        pyart.io.cfradial._create_ncvar(dic, dset, 'foo', ('x'))

        assert 'add_offset' in dic
        assert 'scale_factor' in dic
        assert '_FillValue' in dic

        foo = dset.variables['foo']
        assert foo.dtype == np.dtype('uint8')
        assert_almost_equal(foo.scale_factor, 0.5)
        assert_almost_equal(foo.add_offset, 100)
        assert foo[0] is np.ma.masked
        assert_almost_equal(foo[1], 100.5)
        assert_almost_equal(foo[-1], 227.5)

        dset.close()


def test_calculate_scale_and_offset():
    data = np.arange(128, dtype='float32')
    scale, offset, fill = pyart.io.cfradial._calculate_scale_and_offset(
        {'data': data}, np.dtype('u1'))

    assert_almost_equal(scale, 0.5)
    assert_almost_equal(offset, -0.5)
    assert_almost_equal(fill, 0)

    # check that nans and infs are ignored
    data[5] = np.nan
    data[6] = np.inf
    scale, offset, fill = pyart.io.cfradial._calculate_scale_and_offset(
        {'data': data}, np.dtype('u1'))
    assert_almost_equal(scale, 0.5)
    assert_almost_equal(offset, -0.5)
    assert_almost_equal(fill, 0)

    # maximum < minimum raises ValueError
    pytest.raises(ValueError, pyart.io.cfradial._calculate_scale_and_offset,
                  {'data': data}, np.dtype('u1'), 100., 99)

    # maximum == minimum issues warning
    with warnings.catch_warnings(record=True) as w:
        assert len(w) == 0
        pyart.io.cfradial._calculate_scale_and_offset(
            {'data': data}, np.dtype('u1'), 100, 100)
        assert len(w) == 1



"""
Functions for writing UF files.

"""

import math
import struct
import warnings

import numpy as np
from netCDF4 import num2date

from ..config import get_field_mapping
from .uf import _LIGHT_SPEED
from .uffile import UF_MANDATORY_HEADER
from .uffile import UF_OPTIONAL_HEADER
from .uffile import UF_DATA_HEADER
from .uffile import UF_FIELD_POSITION
from .uffile import UF_FIELD_HEADER
from .uffile import UF_FSI_VEL
from .uffile import POLARIZATION_STR


def write_uf(filename, radar, uf_field_names=None, radar_field_names=False,
             exclude_fields=None, field_write_order=None, volume_start=None,
             templates_extra=None):
    """
    Write a Radar object to a UF file.

    Create a UF file containing data from the provided radar instance.
    The UF file will contain instrument parameters from the following
    dictionaries if they contained in radar.instrument_parameters:

        * radar_beam_width_h
        * radar_beam_width_v
        * radar_receiver_bandwidth
        * frequency
        * pulse_width
        * prt
        * polarization_mode
        * nyquist_velocity

    If any of these parameter are not present a default or sentinel value
    will be written in the UF file in the place of the parameter. This is also
    true for the data in the scan_rate attribute.

    Radar fields will be scaled and rounded to integer values when writing to
    UF files. The scale factor for each field can be specified in the
    `_UF_scale_factor` key for each field dictionary. If not specified the
    default scaling (100) will be used.

    Parameters
    ----------
    filename : str or file-like object.
        Filename of UF file to create. If a file-like object is specified
        data will be written using the write method.
    radar : Radar
        Radar object from which to create UF file.
    uf_field_names : dict or None, optional
        Mapping between radar fields and two character UF data type names.
        Field names mapped to None or with no mapping will be excluded from
        writing. If None, the default mappings for UF files will be used.
    radar_field_names : bool, optional
        True to use the radar field names as the field names of the UF
        fields. False to use the uf_field_names mapping to generate UF field
        names. The `exclude_fields` argument can still be used to exclude
        fields from the UF file when this parameter is True. When reading a UF
        file using `file_field_names=True` set this parameter to True to write
        a UF file with the same field names.
    exclude_fields : list or None, optional
        List of radar fields to exclude from writing.
    field_write_order : list or None, optional
        Order in which radar fields should be written out in the UF file.
        None, the default, will determine a valid order automatically.
    volume_start : datetime, optional
        Start of volume used to set UF volume structure elements.
    templates_extra : dict of dict or None
        Advanced usage parameter for setting UF structure templates.
        Elements defined in dictionaries with keys 'mandatory_header',
        'optional_header', and 'field_header' will be used to build the
        structure template.

    """
    if hasattr(filename, 'write'):
        fhandle = filename
        close = False
    else:
        fhandle = open(filename, 'wb')
        close = True

    field_mapping = _find_field_mapping(
        radar, uf_field_names, radar_field_names, exclude_fields)

    if field_write_order is None:
        field_write_order = list(field_mapping.keys())

    raycreator = UFRayCreator(
        radar, field_mapping, field_write_order, volume_start=volume_start,
        templates_extra=templates_extra)

    for ray_num in range(radar.nrays):

        ray_bytes = raycreator.make_ray(ray_num)
        pad = struct.pack(b'>i', raycreator.record_length * 2)

        fhandle.write(pad)
        fhandle.write(ray_bytes)
        fhandle.write(pad)

    if close:
        fhandle.close()
    return


def _find_field_mapping(
        radar, uf_field_names, radar_field_names, exclude_fields):
    """ Return a dictionary mapping radar fields to UF data types. """
    if uf_field_names is None:
        uf_field_names = get_field_mapping('write_uf')
    if exclude_fields is None:
        exclude_fields = []

    field_mapping = {}
    for radar_field in radar.fields.keys():

        if radar_field in exclude_fields:
            continue

        if radar_field_names:
            data_type = radar_field
        else:
            if radar_field not in uf_field_names:
                continue
            data_type = uf_field_names[radar_field]

        if data_type is None:
            continue
        field_mapping[radar_field] = data_type
    return field_mapping


class UFRayCreator(object):
    """
    A class for generating UF rays for writing UF file.

    Parameters
    ----------
    radar : Radar
        Radar used to create rays.
    field_write_order : list
        Order in which radar fields should be written out in the UF file.
        None, the default, will determine a valid order automatically.
    volume_start : datetime, optional
        Start of volume used to set UF volume fields.
    templates_extra : dict of dict, optional
        Advanced usage parameter for setting UF structure templates.
        Elements defined in dictionaries with keys 'mandatory_header',
        'optional_header', and 'field_header' will be added to the
        appropriate structure template.

    """

    def __init__(self, radar, field_mapping, field_write_order,
                 volume_start=None, templates_extra=None):
        """ Initialize the object. """
        self.radar = radar
        self.field_mapping = field_mapping
        self.field_write_order = field_write_order
        self.record_length = self._calc_record_length(
            radar, field_mapping, field_write_order)
        self.ray_num_to_sweep_num = self._calc_ray_num_to_sweep_num(radar)

        self.mandatory_header_template = UF_MANDATORY_HEADER_TEMPLATE.copy()
        self.optional_header_template = UF_OPTIONAL_HEADER_TEMPLATE.copy()
        self.field_header_template = UF_FIELD_HEADER_TEMPLATE.copy()

        self._set_mandatory_header_location()
        self._set_optional_header_time(volume_start)
        self._set_field_header()
        if templates_extra is not None:
            self._parse_custom_templates(templates_extra)
        return

    @staticmethod
    def _calc_ray_num_to_sweep_num(radar):
        """ Return an array mapping ray number to sweep numbers. """
        ray_num_to_sweep_num = np.zeros((radar.nrays, ), dtype='int32')
        for isweep, sweep_slice in enumerate(radar.iter_slice()):
            ray_num_to_sweep_num[sweep_slice] = isweep
        return ray_num_to_sweep_num

    @staticmethod
    def _calc_record_length(radar, field_mapping, field_write_order):
        """ Return the record length in 2-byte words. """
        # record length is given by the sum of
        # 45 word (90 byte) mandatory header
        # 14 word (28 byte) optional header which is always written
        # 3 word (6 byte) data header
        # For each field:
        #   * nbin words for the data (stored as int16)
        #   * 2 words for the data type name and offset in the data header
        #   * 19 words for the field header
        # For each velocity-like field:
        #   * 2 words (4-bytes) for the FSI velocity structure
        nfields = len(field_write_order)
        data_types = [field_mapping[field].encode('ascii') for
                      field in field_write_order]
        nvel = sum(
            [data_type in UF_VEL_DATA_TYPES for data_type in data_types])
        return 45+14+3 + (radar.ngates+2+19)*nfields + 2*nvel

    def _set_optional_header_time(self, volume_start):
        """ Populate the optional header template with the volume start. """
        header = self.optional_header_template
        if volume_start is None:
            volume_start = num2date(self.radar.time['data'][0],
                                    self.radar.time['units'])
        header['volume_hour'] = volume_start.hour
        header['volume_minute'] = volume_start.minute
        header['volume_second'] = volume_start.second

    def _set_mandatory_header_location(self):
        """ Populate the mandatory header template with the location. """
        header = self.mandatory_header_template

        degrees, minutes, seconds = _d_to_dms(self.radar.latitude['data'][0])
        header['latitude_degrees'] = int(degrees)
        header['latitude_minutes'] = int(minutes)
        header['latitude_seconds'] = int(seconds * 64)

        degrees, minutes, seconds = _d_to_dms(self.radar.longitude['data'][0])
        header['longitude_degrees'] = int(degrees)
        header['longitude_minutes'] = int(minutes)
        header['longitude_seconds'] = int(seconds * 64)

        header['height_above_sea_level'] = int(self.radar.altitude['data'][0])
        return

    def _set_field_header(self):
        """ Populate the field header template with radar parameters. """
        # these parameter are assumed to be constant for all rays in
        # the volume, UF can store volumes where these change between
        # rays but Py-ART does not support writing such volumes
        header = self.field_header_template

        range_step = self.radar.range['meters_between_gates']
        range_start = self.radar.range['meters_to_center_of_first_gate']
        range_start -= (range_step / 2.)  # range bin center to edge
        range_start_km, range_start_m = divmod(range_start, 1000)
        header['range_start_km'] = int(range_start_km)
        header['range_start_m'] = int(round(range_start_m))
        header['range_spacing_m'] = int(round(range_step))

        iparams = self.radar.instrument_parameters

        if iparams is not None and 'radar_beam_width_h' in iparams:
            beam_width_h = iparams['radar_beam_width_h']['data'][0]
            header['beam_width_h'] = int(round(beam_width_h * 64))
        else:
            header['beam_width_h'] = UF_MISSING_VALUE

        if iparams is not None and 'radar_beam_width_v' in iparams:
            beam_width_v = iparams['radar_beam_width_v']['data'][0]
            header['beam_width_v'] = int(round(beam_width_v * 64))
        else:
            header['beam_width_v'] = UF_MISSING_VALUE

        if iparams is not None and 'radar_receiver_bandwidth' in iparams:
            bandwidth = iparams['radar_receiver_bandwidth']['data'][0]
            header['bandwidth'] = int(round(bandwidth / 1.e6 * 16))
        else:
            header['bandwidth'] = UF_MISSING_VALUE

        if iparams is not None and 'frequency' in iparams:
            frequency = iparams['frequency']['data'][0]
            header['wavelength_cm'] = int(round(
                _LIGHT_SPEED / frequency * 100 * 64))
        else:
            header['wavelength_cm'] = UF_MISSING_VALUE

        return

    def _parse_custom_templates(self, templates_extra):
        """ Set additional template parameter using provided dictionary. """

        if 'mandatory_header' in templates_extra:
            for key, value in templates_extra['mandatory_header'].items():
                self.mandatory_header_template[key] = value

        if 'optional_header' in templates_extra:
            for key, value in templates_extra['optional_header'].items():
                self.optional_header_template[key] = value

        if 'field_header' in templates_extra:
            for key, value in templates_extra['field_header'].items():
                self.field_header_template[key] = value

        return

    def make_ray(self, ray_num):
        """ Return a byte string representing a complete UF ray. """
        ray = self.make_mandatory_header(ray_num)
        ray += self.make_optional_header()
        ray += self.make_data_header()
        field_positions = self.make_field_position_list()
        ray += self.make_field_position()

        for field_info in field_positions:

            data_type = field_info['data_type']
            offset = field_info['offset_field_header'] + 19
            radar_field = field_info['radar_field']
            if '_UF_scale_factor' in self.radar.fields[radar_field]:
                scale = self.radar.fields[radar_field]['_UF_scale_factor']
            else:
                scale = UF_DEFAULT_SCALE_FACTOR

            if data_type in UF_VEL_DATA_TYPES:
                offset += 2
                vel_header = self.make_fsi_vel(ray_num, scale)
            else:
                vel_header = b''

            field_header = self.make_field_header(offset, ray_num, scale)
            data_array = self.make_data_array(radar_field, ray_num, scale)

            ray += field_header
            ray += vel_header
            ray += data_array.tostring()

        return ray

    def make_mandatory_header(self, ray_num):
        """ Return a byte string representing a UF mandatory header. """

        # time parameters
        ray_time = num2date(self.radar.time['data'][ray_num],
                            self.radar.time['units'])
        header = self.mandatory_header_template
        header['year'] = ray_time.year - 2000
        header['month'] = ray_time.month
        header['day'] = ray_time.day
        header['hour'] = ray_time.hour
        header['minute'] = ray_time.minute
        header['second'] = ray_time.second

        # ray/sweep numbers
        sweep_num = self.ray_num_to_sweep_num[ray_num]
        header['record_number'] = header['ray_number'] = ray_num + 1
        header['sweep_number'] = sweep_num + 1

        # pointing
        azimuth = self.radar.azimuth['data'][ray_num]
        header['azimuth'] = int(round(azimuth * 64))

        elevation = self.radar.elevation['data'][ray_num]
        header['elevation'] = int(round(elevation * 64))

        fixed_angle = self.radar.fixed_angle['data'][sweep_num]
        header['fixed_angle'] = int(round(fixed_angle * 64))

        if self.radar.scan_rate is not None:
            scan_rate = self.radar.scan_rate['data'][ray_num]
        else:
            scan_rate = UF_MISSING_VALUE / 64
        header['sweep_rate'] = int(round(scan_rate * 64))

        if self.radar.scan_type in UF_SWEEP_MODES:
            sweep_mode_number = UF_SWEEP_MODES[self.radar.scan_type]
        else:
            warnings.warn(
                'Unknown scan_type: %s, defaulting to PPI' %
                (self.radar.scan_type))
            sweep_mode_number = UF_SWEEP_MODES['ppi']
        header['sweep_mode'] = sweep_mode_number

        header['record_length'] = self.record_length

        return _pack_structure(header, UF_MANDATORY_HEADER)

    def make_optional_header(self):
        """ Return a byte string representing a UF optional header. """
        header = self.optional_header_template
        return _pack_structure(header, UF_OPTIONAL_HEADER)

    def make_data_header(self):
        """ Return a byte string representing a UF data header. """
        header = UF_DATA_HEADER_TEMPLATE.copy()
        nfields = len(self.field_write_order)
        header['ray_nfields'] = nfields
        header['record_nfields'] = nfields
        return _pack_structure(header, UF_DATA_HEADER)

    def make_field_position_list(self):
        """ Return a list of field position dictionaries. """
        # 62 words (124 bytes) are occupied by the:
        # * mandatory header (90 bytes)
        # * optional header (28)
        # * data header (6)
        # This is followed by 2 words (4 bytes) for each field name/offset
        # Finally one must be added since the offset has origin of 1
        offset = 62 + len(self.field_write_order) * 2 + 1
        field_positions = []
        for radar_field in self.field_write_order:
            data_type = self.field_mapping[radar_field].encode('ascii')
            field_position = UF_FIELD_POSITION_TEMPLATE.copy()
            field_position['data_type'] = data_type
            field_position['offset_field_header'] = offset
            field_position['radar_field'] = radar_field
            field_positions.append(field_position)

            # account for field header and data
            offset += self.radar.ngates + 19
            if data_type in UF_VEL_DATA_TYPES:
                offset += 2  # account for the FSI_VEL structure
        return field_positions

    def make_field_position(self):
        """ Return a byte string representing the UF field positions. """
        fps = self.make_field_position_list()
        return b''.join([_pack_structure(fp, UF_FIELD_POSITION) for fp in fps])

    def make_field_header(self, data_offset, ray_num, scale_factor):
        """ Return a byte string representing a field header. """
        field_header = self.field_header_template
        field_header['nbins'] = self.radar.ngates
        field_header['data_offset'] = data_offset
        field_header['scale_factor'] = scale_factor

        iparams = self.radar.instrument_parameters
        if iparams is not None and 'pulse_width' in iparams:
            pulse_width = iparams['pulse_width']['data'][ray_num]
            field_header['pulse_width_m'] = int(round(
                pulse_width * _LIGHT_SPEED))
        else:
            field_header['pulse_width_m'] = UF_MISSING_VALUE

        if iparams is not None and 'prt' in iparams:
            prt = iparams['prt']['data'][ray_num]
            field_header['prt_ms'] = int(round(prt * 1.e6))
        else:
            field_header['prt_ms'] = UF_MISSING_VALUE

        field_header['polarization'] = 1  # default to horizontal polarization
        sweep_num = self.ray_num_to_sweep_num[ray_num]
        if iparams is not None and 'polarization_mode' in iparams:
            mode = str(iparams['polarization_mode']['data'][sweep_num])
            if mode in POLARIZATION_STR:
                field_header['polarization'] = POLARIZATION_STR.index(mode)
        return _pack_structure(field_header, UF_FIELD_HEADER)

    def make_fsi_vel(self, ray_num, scale):
        """ Return a byte string representing a UF FSI velocity structure. """
        fsi_vel = UF_FSI_VEL_TEMPLATE.copy()
        iparams = self.radar.instrument_parameters
        if iparams is not None and 'nyquist_velocity' in iparams:
            nyquist = iparams['nyquist_velocity']['data'][ray_num]
            fsi_vel['nyquist'] = int(round(nyquist * scale))
        else:
            fsi_vel['nyquist'] = UF_MISSING_VALUE
        return _pack_structure(fsi_vel, UF_FSI_VEL)

    def make_data_array(self, field, ray_num, scale=100.):
        """ Return an array of UF field data. """
        field_data = np.round(self.radar.fields[field]['data'][ray_num]*scale)
        return field_data.filled(-32768).astype('>i2')


def _d_to_dms(in_deg):
    """ Degrees to degree, minutes, seconds. """
    # add or subtract a fraction of a second to fix round off issues
    epsilon = 0.01 / 3600.
    in_deg += epsilon * np.sign(in_deg)
    remain, degrees = math.modf(in_deg)
    remain, minutes = math.modf(remain * 60.)
    remain, seconds = math.modf(remain * 60.)
    return degrees, minutes, seconds


def _pack_structure(dic, structure):
    """ Pack a structure from a dictionary. """
    fmt = '>' + ''.join([i[1] for i in structure])  # UF is big-endian
    values = [dic[i[0]] for i in structure]
    # cast to string as Python 2.7 pack does not except unicode before 2.7.7
    fmt = str(fmt)
    return struct.pack(fmt, *values)


# Constants
UF_MISSING_VALUE = -32768
UF_DEFAULT_SCALE_FACTOR = 100   # default field scale factor
UF_SWEEP_MODES = {
    'calibration': 0,
    'ppi': 1,
    'coplane': 2,
    'rhi': 3,
    'vpt': 4,
    'target': 5,
    'manual': 6,
    'idle': 7,
}
UF_VEL_DATA_TYPES = [b'VF', b'VE', b'VR', b'VT', b'VP']

# Structure dictionary templates
UF_MANDATORY_HEADER_TEMPLATE = {
    'uf_string': b'UF',
    'record_length': 999,
    'offset_optional_header': 46,   # Always include the optional read
    'offset_local_use_header': 60,  # Never include a local use header
    'offset_data_header': 60,       # Data header follows optional header
    'record_number': 999,
    'volume_number': 1,             # always a single volume
    'ray_number': 999,
    'ray_record_number': 1,         # always one recorded per ray
    'sweep_number': 999,
    'radar_name': b'XXXXXXXX',
    'site_name': b'XXXXXXXX',
    'latitude_degrees': 999,
    'latitude_minutes': 999,
    'latitude_seconds': 999,
    'longitude_degrees': 999,
    'longitude_minutes': 999,
    'longitude_seconds': 999,
    'height_above_sea_level': 999,
    'year': 999,
    'month': 999,
    'day': 999,
    'hour': 999,
    'minute': 999,
    'second': 999,
    'time_zone': b'UT',
    'azimuth': 999,
    'elevation': 999,
    'sweep_mode': 999,
    'fixed_angle': 999,
    'sweep_rate': 999,
    'generation_year': 999,
    'generation_month': 999,
    'generation_day': 999,
    'generation_facility_name': b'XXXXXXXX',
    'missing_data_value': UF_MISSING_VALUE,     # standard missing value
}

UF_OPTIONAL_HEADER_TEMPLATE = {
    'project_name': b'XXXXXXXX',
    'baseline_azimuth': UF_MISSING_VALUE,
    'baseline_elevation': UF_MISSING_VALUE,
    'volume_hour': 999,
    'volume_minute': 999,
    'volume_second': 999,
    'tape_name': b'XXXXXXXX',
    'flag': 2,   # default used by RSL
}

UF_DATA_HEADER_TEMPLATE = {
    'ray_nfields': 999,
    'ray_nrecords': 1,  # 1 record per ray
    'record_nfields': 999,
}

UF_FIELD_POSITION_TEMPLATE = {
    'data_type': b'XX',
    'offset_field_header': 999,
}

UF_FIELD_HEADER_TEMPLATE = {
    'data_offset': 999,
    'scale_factor': 999,
    'range_start_km': 999,
    'range_start_m': 999,
    'range_spacing_m': 999,
    'nbins': 999,
    'pulse_width_m': 999,
    'beam_width_h': 999,
    'beam_width_v': 999,
    'bandwidth': 999,
    'polarization': 999,
    'wavelength_cm': 999,
    'sample_size': 90,          # Apparently defaults to 90?
    'threshold_data': b'  ',    # No thresholding field
    'threshold_value': UF_MISSING_VALUE,
    'scale': UF_MISSING_VALUE,
    'edit_code': b'  ',         # unknown use, typically blank or null
    'prt_ms': 999,
    'bits_per_bin': 16,         # 16 bits (2 bytes, 1 word) per bin or gate
}

UF_FSI_VEL_TEMPLATE = {
    'nyquist': 999,
    'spare': 1,     # 1 if commonly used for this unused element
}



"""
Dealias using multidimensional phase unwrapping algorithms.

"""

import numpy as np

from ..config import get_metadata, get_fillvalue
from ._common_dealias import _parse_fields, _parse_gatefilter, _set_limits
from ._common_dealias import _parse_rays_wrap_around, _parse_nyquist_vel

from ._unwrap_1d import unwrap_1d
from ._unwrap_2d import unwrap_2d
from ._unwrap_3d import unwrap_3d


def dealias_unwrap_phase(
        radar, unwrap_unit='sweep', nyquist_vel=None,
        check_nyquist_uniform=True, gatefilter=False,
        rays_wrap_around=None, keep_original=False, set_limits=True,
        vel_field=None, corr_vel_field=None, skip_checks=False, **kwargs):
    """
    Dealias Doppler velocities using multi-dimensional phase unwrapping
    [1]_ and [2]_.

    Parameters
    ----------
    radar : Radar
        Radar object containing Doppler velocities to dealias.
    unwrap_unit : {'ray', 'sweep', 'volume'}, optional
        Unit to unwrap independently. 'ray' will unwrap each ray
        individually, 'sweep' each sweep, and 'volume' will unwrap the entire
        volume in a single pass. 'sweep', the default, often gives superior
        results when the lower sweeps of the radar volume are contaminated by
        clutter. 'ray' does not use the gatefilter parameter and rays where
        gates ared masked will result in poor dealiasing for that ray.
    nyquist_velocity : array like or float, optional
        Nyquist velocity in unit identical to those stored in the radar's
        velocity field, either for each sweep or a single value which will be
        used for all sweeps. None will attempt to determine this value from
        the Radar object. The Nyquist velocity of the first sweep is used
        for all dealiasing unless the unwrap_unit is 'sweep' when the
        velocities of each sweep are used.
    check_nyquist_uniform : bool, optional
        True to check if the Nyquist velocities are uniform for all rays
        within a sweep, False will skip this check. This parameter is ignored
        when the nyquist_velocity parameter is not None.
    gatefilter : GateFilter, None or False, optional.
        A GateFilter instance which specified which gates should be
        ignored when performing de-aliasing. A value of None created this
        filter from the radar moments using any additional arguments by
        passing them to :py:func:`moment_based_gate_filter`. False, the
        default, disables filtering including all gates in the dealiasing.
    rays_wrap_around : bool or None, optional
        True when the rays at the beginning of the sweep and end of the sweep
        should be interpreted as connected when de-aliasing (PPI scans).
        False if they edges should not be interpreted as connected (other scan
        types). None will determine the correct value from the radar
        scan type.
    keep_original : bool, optional
        True to retain the original Doppler velocity values at gates
        where the dealiasing procedure fails or was not applied. False
        does not replacement and these gates will be masked in the corrected
        velocity field.
    set_limits : bool, optional
        True to set valid_min and valid_max elements in the returned
        dictionary. False will not set these dictionary elements.
    vel_field : str, optional
        Field in radar to use as the Doppler velocities during dealiasing.
        None will use the default field name from the Py-ART configuration
        file.
    corr_vel_field : str, optional
        Name to use for the dealiased Doppler velocity field metadata.  None
        will use the default field name from the Py-ART configuration file.
    skip_checks : bool
        True to skip checks verifing that an appropiate unwrap_unit is
        selected, False retains these checked. Setting this parameter to True
        is not recommended and is only offered as an option for extreme cases.

    Returns
    -------
    corr_vel : dict
        Field dictionary containing dealiased Doppler velocities.  Dealiased
        array is stored under the 'data' key.

    References
    ----------
    .. [1] Miguel Arevallilo Herraez, David R. Burton, Michael J. Lalor,
        and Munther A. Gdeisat, "Fast two-dimensional phase-unwrapping
        algorithm based on sorting by reliability following a noncontinuous
        path", Journal Applied Optics, Vol. 41, No. 35 (2002) 7437,
    .. [2] Abdul-Rahman, H., Gdeisat, M., Burton, D., & Lalor, M., "Fast
        three-dimensional phase-unwrapping algorithm based on sorting by
        reliability following a non-continuous path. In W. Osten,
        C. Gorecki, & E. L. Novak (Eds.), Optical Metrology (2005) 32--40,
        International Society for Optics and Photonics.

    """
    vel_field, corr_vel_field = _parse_fields(vel_field, corr_vel_field)
    gatefilter = _parse_gatefilter(gatefilter, radar, **kwargs)
    rays_wrap_around = _parse_rays_wrap_around(rays_wrap_around, radar)
    nyquist_vel = _parse_nyquist_vel(nyquist_vel, radar, check_nyquist_uniform)
    if not skip_checks:
        _verify_unwrap_unit(radar, unwrap_unit)

    # exclude masked and invalid velocity gates
    gatefilter.exclude_masked(vel_field)
    gatefilter.exclude_invalid(vel_field)
    gfilter = gatefilter.gate_excluded

    # raw vel. data possibly with masking
    raw_vdata = radar.fields[vel_field]['data']
    vdata = raw_vdata.view(np.ndarray)      # mask removed

    # perform dealiasing
    if unwrap_unit == 'ray':
        # 1D unwrapping does not use the gate filter nor respect
        # masked gates in the rays.  No information from the radar object is
        # needed for the unfolding
        data = _dealias_unwrap_1d(vdata, nyquist_vel)
    elif unwrap_unit == 'sweep':
        data = _dealias_unwrap_2d(
            radar, vdata, nyquist_vel, gfilter, rays_wrap_around)
    elif unwrap_unit == 'volume':
        data = _dealias_unwrap_3d(
            radar, vdata, nyquist_vel, gfilter, rays_wrap_around)
    else:
        message = ("Unknown `unwrap_unit` parameter, must be one of "
                   "'ray', 'sweep', or 'volume'.")
        raise ValueError(message)

    # fill_value from the velocity dictionary if present
    fill_value = radar.fields[vel_field].get(
        '_FillValue', get_fillvalue())

    # mask filtered gates
    if np.any(gfilter):
        data = np.ma.array(data, mask=gfilter, fill_value=fill_value)

    # restore original values where dealiasing not applied
    if keep_original:
        data[gfilter] = raw_vdata[gfilter]

    # return field dictionary containing dealiased Doppler velocities
    corr_vel = get_metadata(corr_vel_field)
    corr_vel['data'] = data
    corr_vel['_FillValue'] = fill_value

    if set_limits:
        # set valid_min and valid_max in corr_vel
        _set_limits(data, nyquist_vel, corr_vel)
    return corr_vel


def _dealias_unwrap_3d(radar, vdata, nyquist_vel, gfilter, rays_wrap_around):
    """ Dealias using 3D phase unwrapping (full volume at once). """

    # form cube and scale to phase units
    nyquist_vel = nyquist_vel[0]   # must be uniform, not checked
    shape = (radar.nsweeps, -1, radar.ngates)
    scaled_cube = (np.pi * vdata / nyquist_vel).reshape(shape)
    filter_cube = gfilter.reshape(shape)

    # perform unwrapping
    wrapped = np.require(scaled_cube, np.float64, ['C'])
    mask = np.require(filter_cube, np.uint8, ['C'])
    unwrapped = np.empty_like(wrapped, dtype=np.float64, order='C')
    unwrap_3d(wrapped, mask, unwrapped, [False, rays_wrap_around, False])

    # scale back to velocity units
    unwrapped_cube = unwrapped * nyquist_vel / np.pi
    unwrapped_volume = unwrapped_cube.reshape(-1, radar.ngates)
    unwrapped_volume = unwrapped_volume.astype(vdata.dtype)
    return unwrapped_volume


def _dealias_unwrap_1d(vdata, nyquist_vel):
    """ Dealias using 1D phase unwrapping (ray-by-ray). """
    # nyquist_vel is only available sweep by sweep which has been lost at
    # this point.  Metioned in the documentation
    nyquist_vel = nyquist_vel[0]
    data = np.empty_like(vdata)
    for i, ray in enumerate(vdata):
        # extract ray and scale to phase units
        scaled_ray = ray * np.pi / nyquist_vel

        # perform unwrapping
        wrapped = np.require(scaled_ray, np.float64, ['C'])
        unwrapped = np.empty_like(wrapped, dtype=np.float64, order='C')
        unwrap_1d(wrapped, unwrapped)

        # scale back into velocity units and store
        data[i] = unwrapped * nyquist_vel / np.pi
    return data


def _dealias_unwrap_2d(radar, vdata, nyquist_vel, gfilter, rays_wrap_around):
    """ Dealias using 2D phase unwrapping (sweep-by-sweep). """
    data = np.zeros_like(vdata)
    for nsweep, sweep_slice in enumerate(radar.iter_slice()):
        # extract sweep and scale to phase units
        sweep_nyquist_vel = nyquist_vel[nsweep]
        scaled_sweep = vdata[sweep_slice] * np.pi / sweep_nyquist_vel
        sweep_mask = gfilter[sweep_slice]

        # perform unwrapping
        wrapped = np.require(scaled_sweep, np.float64, ['C'])
        mask = np.require(sweep_mask, np.uint8, ['C'])
        unwrapped = np.empty_like(wrapped, dtype=np.float64, order='C')
        unwrap_2d(wrapped, mask, unwrapped, [rays_wrap_around, False])

        # scale back into velocity units and store
        data[sweep_slice, :] = unwrapped * sweep_nyquist_vel / np.pi
    return data


def _verify_unwrap_unit(radar, unwrap_unit):
    """
    Verify that the radar supports the requested unwrap unit

    raises a ValueError if the unwrap_unit is not supported.
    """
    if unwrap_unit == 'sweep' or unwrap_unit == 'volume':
        if _is_radar_sequential(radar) is False:
            mess = ("rays are not sequentially ordered, must use 'ray' "
                    "unwrap_unit.")
            raise ValueError(mess)
    if unwrap_unit == 'volume':
        if _is_radar_cubic(radar) is False:
            mess = "Non-cubic radar volume, 'volume' unwrap_unit invalid. "
            raise ValueError(mess)
        if _is_radar_sweep_aligned(radar) is False:
            mess = ("Angle in sequential sweeps in radar volumes are not "
                    "aligned, 'volume unwrap_unit invalid.")
            raise ValueError(mess)


def _is_radar_cubic(radar):
    """ Test if a radar is cubic (sweeps have the same number of rays). """
    rays_per_sweep = radar.rays_per_sweep['data']
    return bool(np.all(rays_per_sweep == rays_per_sweep[0]))


def _is_radar_sweep_aligned(radar, diff=0.1):
    """
    Test that all sweeps in the radar sample nearly the same angles.

    Test that the maximum difference in sweep sampled angles is below
    `diff` degrees. The radar should first be tested to verify that is cubic
    before calling this function using the _is_radar_cubic function.

    """
    if radar.nsweeps == 1:
        return True     # all single sweep volume are sweep aligned
    if radar.scan_type == 'ppi':
        angles = radar.azimuth['data']
    elif radar.scan_type == 'rhi':
        angles = radar.elevation['data']
    else:
        raise ValueError('invalid scan_type: %s' % (radar.scan_type))
    starts = radar.sweep_start_ray_index['data']
    ends = radar.sweep_end_ray_index['data']
    ref_angles = angles[starts[0]:ends[0] + 1]
    for start, end in zip(starts, ends):
        test_angles = angles[start:end+1]
        if np.any(np.abs(test_angles - ref_angles) > diff):
            return False
    return True


def _is_radar_sequential(radar):
    """ Test if all sweeps in radar are sequentially ordered. """
    for i in range(radar.nsweeps):
        if not _is_sweep_sequential(radar, i):
            return False
    return True


def _is_sweep_sequential(radar, sweep_number):
    """ Test if a specific sweep is sequentially ordered. """
    start = radar.sweep_start_ray_index['data'][sweep_number]
    end = radar.sweep_end_ray_index['data'][sweep_number]
    if radar.scan_type == 'ppi':
        angles = radar.azimuth['data'][start:end+1]
    elif radar.scan_type == 'rhi':
        angles = radar.elevation['data'][start:end+1]
    elif radar.scan_type == 'vpt':
        # for VPT scan time should not run backwards, so time is the
        # equivalent variable to an angle.
        angles = radar.time['data']
    else:
        raise ValueError('invalid scan_type: %s' % (radar.scan_type))
    rolled_angles = np.roll(angles, -np.argmin(angles))
    return np.all(np.diff(rolled_angles) >= 0)



""" Unit tests for spectra_calculations.py """

from numpy.testing import assert_allclose
import pytest

from pyart.retrieve import spectra_moments
from pyart.testing import make_target_spectra_radar

try:
    import xarray as xr
    _XARRAY_AVAILABLE = True
except ImportError:
    _XARRAY_AVAILABLE = False

@pytest.mark.skipif(not _XARRAY_AVAILABLE,
                    reason="Xarray is not installed.")
def test_spectra_moments():
    radar = make_target_spectra_radar()
    fields = spectra_moments(radar)
    assert len(fields.keys()) == 5
    assert_allclose(
        fields['reflectivity']['data'][0, 0:5],
        [71.34390136, 71.34390136, 71.34390136, 71.34390136, 71.34390136],
        atol=1e-14)
    assert_allclose(
        fields['velocity']['data'][0, 0:5],
        [2.73382043e-16, 2.73382043e-16, 2.73382043e-16, 2.73382043e-16,
         2.73382043e-16],
        atol=1e-14)
    assert_allclose(
        fields['spectrum_width']['data'][0, 0:5],
        [2.85047687, 2.85047687, 2.85047687, 2.85047687, 2.85047687],
        atol=1e-14)
    assert_allclose(
        fields['skewness']['data'][0, 0:5],
        [9.44294386e-17, 9.44294386e-17, 9.44294386e-17, 9.44294386e-17,
         9.44294386e-17],
        atol=1e-14)
    assert_allclose(
        fields['kurtosis']['data'][0, 0:5],
        [2.79911197, 2.79911197, 2.79911197, 2.79911197, 2.79911197],
        atol=1e-14)



"""
An class for holding gridded Radar data.

"""

import numpy as np
from netCDF4 import num2date

try:
    import xarray
    _XARRAY_AVAILABLE = True
except ImportError:
    _XARRAY_AVAILABLE = False

try:
    import pyproj
    _PYPROJ_AVAILABLE = True
except ImportError:
    _PYPROJ_AVAILABLE = False

from ..config import get_metadata
from ..exceptions import MissingOptionalDependency
from ..lazydict import LazyLoadDict
from .transforms import cartesian_to_geographic
from .transforms import cartesian_vectors_to_geographic


class Grid(object):
    """
    A class for storing rectilinear gridded radar data in Cartesian coordinate.

    Refer to the attribute section for information on the parameters.

    To create a Grid object using legacy parameters present in Py-ART version
    1.5 and before, use :py:func:`from_legacy_parameters`,
    grid = Grid.from_legacy_parameters(fields, axes, metadata).

    Attributes
    ----------
    time : dict
        Time of the grid.
    fields : dict of dicts
        Moments from radars or other variables.
    metadata : dict
        Metadata describing the grid.
    origin_longitude, origin_latitude, origin_altitude : dict
        Geographic coordinate of the origin of the grid.
    x, y, z : dict, 1D
        Distance from the grid origin for each Cartesian coordinate axis in a
        one dimensional array. Defines the spacing along the three grid axes
        which is repeated throughout the grid, making a rectilinear grid.
    nx, ny, nz : int
        Number of grid points along the given Cartesian dimension.
    projection : dic or str
        Projection parameters defining the map projection used to transform
        from Cartesian to geographic coordinates. None will use the default
        dictionary with the 'proj' key set to 'pyart_aeqd' indicating that
        the native Py-ART azimuthal equidistant projection is used. Other
        values should specify a valid pyproj.Proj projparams dictionary or
        string. The special key '_include_lon_0_lat_0' is removed when
        interpreting this dictionary. If this key is present and set to True,
        which is required when proj='pyart_aeqd', then the radar longitude and
        latitude will be added to the dictionary as 'lon_0' and 'lat_0'.
        Use the :py:func:`get_projparams` method to retrieve a copy of this
        attribute dictionary with this special key evaluated.
    radar_longitude, radar_latitude, radar_altitude : dict or None, optional
        Geographic location of the radars which make up the grid.
    radar_time : dict or None, optional
        Start of collection for the radar which make up the grid.
    radar_name : dict or None, optional
        Names of the radars which make up the grid.
    nradar : int
        Number of radars whose data was used to make the grid.
    projection_proj : Proj
        pyproj.Proj instance for the projection specified by the projection
        attribute. If the 'pyart_aeqd' projection is specified accessing this
        attribute will raise a ValueError.
    point_x, point_y, point_z : LazyLoadDict
        The Cartesian locations of all grid points from the origin in the
        three Cartesian coordinates. The three dimensional data arrays
        contained these attributes are calculated from the x, y, and z
        attributes. If these attributes are changed use :py:func:
        `init_point_x_y_z` to reset the attributes.
    point_longitude, point_latitude : LazyLoadDict
        Geographic location of each grid point. The projection parameter(s)
        defined in the `projection` attribute are used to perform an inverse
        map projection from the Cartesian grid point locations relative to
        the grid origin. If these attributes are changed use
        :py:func:`init_point_longitude_latitude` to reset the attributes.
    point_altitude : LazyLoadDict
        The altitude of each grid point as calculated from the altitude of the
        grid origin and the Cartesian z location of each grid point.  If this
        attribute is changed use :py:func:`init_point_altitude` to reset the
        attribute.

    """
    def __init__(self, time, fields, metadata,
                 origin_latitude, origin_longitude, origin_altitude, x, y, z,
                 projection=None, radar_latitude=None, radar_longitude=None,
                 radar_altitude=None, radar_time=None, radar_name=None):
        """ Initalize object. """

        self.time = time
        self.fields = fields
        self.metadata = metadata
        self.origin_latitude = origin_latitude
        self.origin_longitude = origin_longitude
        self.origin_altitude = origin_altitude
        self.x = x
        self.y = y
        self.z = z
        self.nx = len(x['data'])
        self.ny = len(y['data'])
        self.nz = len(z['data'])
        if projection is None:
            self.projection = {
                'proj': 'pyart_aeqd', '_include_lon_0_lat_0': True}
        else:
            self.projection = projection

        self.radar_latitude = radar_latitude
        self.radar_longitude = radar_longitude
        self.radar_altitude = radar_altitude
        self.radar_time = radar_time
        self.radar_name = radar_name
        self.nradar = self._find_and_check_nradar()

        # initialize attributes with Lazy load dictionaries
        self.init_point_x_y_z()
        self.init_point_longitude_latitude()
        self.init_point_altitude()

        return

    def __getstate__(self):
        """ Return object's state which can be pickled. """
        state = self.__dict__.copy()  # copy the objects state
        # Remove unpicklable entries (those which are lazily loaded
        del state['point_x']
        del state['point_y']
        del state['point_z']
        del state['point_latitude']
        del state['point_longitude']
        del state['point_altitude']
        return state

    def __setstate__(self, state):
        """ Restore unpicklable entries from pickled object. """
        self.__dict__.update(state)
        self.init_point_x_y_z()
        self.init_point_longitude_latitude()
        self.init_point_altitude()

    @property
    def projection_proj(self):
        # Proj instance as specified by the projection attribute.
        # Raises a ValueError if the pyart_aeqd projection is specified.
        projparams = self.get_projparams()
        if projparams['proj'] == 'pyart_aeqd':
            raise ValueError(
                'Proj instance can not be made for the pyart_aeqd projection')
        if not _PYPROJ_AVAILABLE:
            raise MissingOptionalDependency(
                "PyProj is required to create a Proj instance but it "
                + "is not installed")
        proj = pyproj.Proj(projparams)
        return proj

    def get_projparams(self):
        """ Return a projparam dict from the projection attribute. """
        projparams = self.projection.copy()
        if projparams.pop('_include_lon_0_lat_0', False):
            projparams['lon_0'] = self.origin_longitude['data'][0]
            projparams['lat_0'] = self.origin_latitude['data'][0]
        return projparams

    def _find_and_check_nradar(self):
        """
        Return the number of radars which were used to create the grid.

        Examine the radar attributes to determine the number of radars which
        were used to create the grid. If the size of the radar attributes
        are inconsistent a ValueError is raised by this method.
        """
        nradar_set = False
        nradar = 0

        if self.radar_latitude is not None:
            nradar = len(self.radar_latitude['data'])
            nradar_set = True

        if self.radar_longitude is not None:
            if nradar_set and len(self.radar_longitude['data']) != nradar:
                raise ValueError("Inconsistent length of radar_ arguments.")
            nradar = len(self.radar_longitude['data'])
            nradar_set = True

        if self.radar_altitude is not None:
            if nradar_set and len(self.radar_altitude['data']) != nradar:
                raise ValueError("Inconsistent length of radar_ arguments.")
            nradar = len(self.radar_altitude['data'])
            nradar_set = True

        if self.radar_time is not None:
            if nradar_set and len(self.radar_time['data']) != nradar:
                raise ValueError("Inconsistent length of radar_ arguments.")
            nradar = len(self.radar_time['data'])
            nradar_set = True

        if self.radar_name is not None:
            if nradar_set and len(self.radar_name['data']) != nradar:
                raise ValueError("Inconsistent length of radar_ arguments.")
            nradar = len(self.radar_name['data'])
            nradar_set = True

        return nradar

    # Attribute init/reset methods
    def init_point_x_y_z(self):
        """ Initialize or reset the point_{x, y, z} attributes. """
        self.point_x = LazyLoadDict(get_metadata('point_x'))
        self.point_x.set_lazy('data', _point_data_factory(self, 'x'))

        self.point_y = LazyLoadDict(get_metadata('point_y'))
        self.point_y.set_lazy('data', _point_data_factory(self, 'y'))

        self.point_z = LazyLoadDict(get_metadata('point_z'))
        self.point_z.set_lazy('data', _point_data_factory(self, 'z'))

    def init_point_longitude_latitude(self):
        """
        Initialize or reset the point_{longitude, latitudes} attributes.
        """
        point_longitude = LazyLoadDict(get_metadata('point_longitude'))
        point_longitude.set_lazy('data', _point_lon_lat_data_factory(self, 0))
        self.point_longitude = point_longitude

        point_latitude = LazyLoadDict(get_metadata('point_latitude'))
        point_latitude.set_lazy('data', _point_lon_lat_data_factory(self, 1))
        self.point_latitude = point_latitude

    def init_point_altitude(self):
        """ Initialize the point_altitude attribute. """
        point_altitude = LazyLoadDict(get_metadata('point_altitude'))
        point_altitude.set_lazy('data', _point_altitude_data_factory(self))
        self.point_altitude = point_altitude

    def write(self, filename, format='NETCDF4', arm_time_variables=False,
              arm_alt_lat_lon_variables=False):
        """
        Write the the Grid object to a NetCDF file.

        Parameters
        ----------
        filename : str
            Filename to save to.
        format : str, optional
            NetCDF format, one of 'NETCDF4', 'NETCDF4_CLASSIC',
            'NETCDF3_CLASSIC' or 'NETCDF3_64BIT'.
        arm_time_variables : bool, optional
            True to write the ARM standard time variables base_time and
            time_offset. False will not write these variables.
        arm_alt_lat_lon_variables : bool, optional
            True to write the ARM standard alt, lat, lon variables.
            False will not write these variables.

        """
        # delayed import to avoid circular import
        from ..io.grid_io import write_grid

        write_grid(filename, self, format=format,
                   arm_time_variables=arm_time_variables,
                   arm_alt_lat_lon_variables=arm_alt_lat_lon_variables)

    def to_xarray(self):
        """
        Convert the Grid object to an xarray format.

        Attributes
        ----------
        time : dict
            Time of the grid.
        fields : dict of dicts
            Moments from radars or other variables.
        longitude, latitude : dict, 2D
            Arrays of latitude and longitude for the grid height level.
        x, y, z : dict, 1D
            Distance from the grid origin for each Cartesian coordinate axis
            in a one dimensional array.
            
        """

        if not _XARRAY_AVAILABLE:
            raise MissingOptionalDependency(
                'Xarray is required to use Grid.to_xarray but is not '
                 + 'installed!')

        lon, lat = self.get_point_longitude_latitude()
        z = self.z['data']
        y = self.y['data']
        x = self.x['data']

        time = np.array([num2date(self.time['data'][0],
                                  self.time['units'])])
        
        ds = xarray.Dataset()
        for field in list(self.fields.keys()):
            field_data = self.fields[field]['data']
            data = xarray.DataArray(np.ma.expand_dims(field_data, 0),
                                    dims=('time', 'z', 'y', 'x'),
                                    coords={'time' : (['time'], time),
                                            'z' : (['z'], z),
                                            'lat' : (['y'], lat[:, 0]),
                                            'lon' : (['x'], lon[0, :]),
                                            'y' : (['y'], y),
                                            'x' : (['x'], x)})
            for meta in list(self.fields[field].keys()):
                if meta is not 'data':
                    data.attrs.update({meta: self.fields[field][meta]})

            ds[field] = data
            ds.lon.attrs = [('long_name', 'longitude of grid cell center'),
                            ('units', 'degree_E'),
                            ('standard_name', 'Longitude')]
            ds.lat.attrs = [('long_name', 'latitude of grid cell center'),
                            ('units', 'degree_N'),
                            ('standard_name', 'Latitude')]

            ds.z.attrs = get_metadata('z')
            ds.y.attrs = get_metadata('y')
            ds.x.attrs = get_metadata('x')
            
            ds.z.encoding['_FillValue'] = None
            ds.lat.encoding['_FillValue'] = None
            ds.lon.encoding['_FillValue'] = None
            ds.close()
        return ds

    def add_field(self, field_name, field_dict, replace_existing=False):
        """
        Add a field to the object.

        Parameters
        ----------
        field_name : str
            Name of the field to the fields dictionary.
        field_dict : dict
            Dictionary containing field data and metadata.
        replace_existing : bool, optional
            True to replace the existing field with key field_name if it
            exists, overwriting the existing data. If False, a ValueError is
            raised if field_name already exists.

        """
        # checks to make sure input field dictionary is valid
        if 'data' not in field_dict:
            raise KeyError('Field dictionary must contain a "data" key')
        if field_name in self.fields and replace_existing is False:
            raise ValueError('A field named %s already exists' % (field_name))
        if field_dict['data'].shape != (self.nz, self.ny, self.nx):
            raise ValueError('Field has invalid shape')

        self.fields[field_name] = field_dict

    def get_point_longitude_latitude(self, level=0, edges=False):
        """
        Return arrays of longitude and latitude for a given grid height level.

        Parameters
        ----------
        level : int, optional
            Grid height level at which to determine latitudes and longitudes.
            This is not currently used as all height level have the same
            layout.
        edges : bool, optional
            True to calculate the latitude and longitudes of the edges by
            interpolating between Cartesian coordinates points and
            extrapolating at the boundaries. False to calculate the locations
            at the centers.

        Returns
        -------
        longitude, latitude : 2D array
            Arrays containing the latitude and longitudes, in degrees, of the
            grid points or edges between grid points for the given height.

        """
        x = self.x['data']
        y = self.y['data']
        projparams = self.get_projparams()
        return cartesian_vectors_to_geographic(x, y, projparams, edges=edges)


def _point_data_factory(grid, coordinate):
    """ Return a function which returns the locations of all points. """
    def _point_data():
        """ The function which returns the locations of all points. """
        reg_x = grid.x['data']
        reg_y = grid.y['data']
        reg_z = grid.z['data']
        if coordinate == 'x':
            return np.tile(reg_x, (len(reg_z), len(reg_y), 1)).swapaxes(2, 2)
        elif coordinate == 'y':
            return np.tile(reg_y, (len(reg_z), len(reg_x), 1)).swapaxes(1, 2)
        else:
            assert coordinate == 'z'
            return np.tile(reg_z, (len(reg_x), len(reg_y), 1)).swapaxes(0, 2)
    return _point_data


def _point_lon_lat_data_factory(grid, coordinate):
    """ Return a function which returns the geographic locations of points. """
    def _point_lon_lat_data():
        """ The function which returns the geographic point locations. """
        x = grid.point_x['data']
        y = grid.point_y['data']
        projparams = grid.get_projparams()
        geographic_coords = cartesian_to_geographic(x, y, projparams)
        # Set point_latitude['data'] when point_longitude['data'] is evaluated
        # and vice-versa.  This ensures that both attributes contain data from
        # the same map projection and that the map projection only needs to be
        # evaluated once.
        if coordinate == 0:
            grid.point_latitude['data'] = geographic_coords[1]
        else:
            grid.point_longitude['data'] = geographic_coords[0]
        return geographic_coords[coordinate]
    return _point_lon_lat_data


def _point_altitude_data_factory(grid):
    """ Return a function which returns the point altitudes. """
    def _point_altitude_data():
        """ The function which returns the point altitudes. """
        return grid.origin_altitude['data'][0] + grid.point_z['data']
    return _point_altitude_data



"""
Retrieval of QVPs from a radar object

"""

import numpy as np

from ..core.transforms import antenna_to_cartesian


def quasi_vertical_profile(radar, desired_angle=None, fields=None, gatefilter=None):

    """
    Quasi Vertical Profile.

    Creates a QVP object containing fields from a radar object that can
    be used to plot and produce the quasi vertical profile

    Parameters
    ----------
    radar : Radar
        Radar object used.
    field : string
        Radar field to use for QVP calculation.
    desired_angle : float
        Radar tilt angle to use for indexing radar field data.
        None will result in wanted_angle = 20.0

    Other Parameters
    ----------------
    gatefilter : GateFilter
        A GateFilter indicating radar gates that should be excluded
        from the import qvp calculation

    Returns
    -------
    qvp : Dictonary
        A quasai vertical profile object containing fields
        from a radar object

    References
    ----------
    Troemel, S., M. Kumjian, A. Ryzhkov, and C. Simmer, 2013: Backscatter
    differential phase - estimation and variability. J Appl. Meteor. Clim..
    52, 2529 - 2548.

    Troemel, S., A. Ryzhkov, P. Zhang, and C. Simmer, 2014: Investigations
    of backscatter differential phase in the melting layer. J. Appl. Meteorol.
    Clim. 54, 2344 - 2359.

    Ryzhkov, A., P. Zhang, H. Reeves, M. Kumjian, T. Tschallener, S. Tromel,
    C. Simmer, 2015: Quasi-vertical profiles - a new way to look at polarimetric
    radar data. Submitted to J. Atmos. Oceanic Technol.

    """

    # Creating an empty dictonary
    qvp = {}

    # Setting the desired radar angle and getting index value for desired radar angle
    if desired_angle is None:
        desired_angle = 20.0
    index = abs(radar.fixed_angle['data'] - desired_angle).argmin()
    radar_slice = radar.get_slice(index)

    # Printing radar tilt angles and radar elevation
    print(radar.fixed_angle['data'])
    print(radar.elevation['data'][-1])

    # Setting field parameters
    # If fields is None then all radar fields pulled else defined field is used
    if fields is None:
        fields = radar.fields

        for field in fields:

            # Filtering data based on defined gatefilter
            # If none is defined goes to else statement
            if gatefilter is not None:
                get_fields = radar.get_field(index, field)
                mask_fields = np.ma.masked_where(gatefilter.gate_excluded[radar_slice],
                                                 get_fields)
                radar_fields = np.ma.mean(mask_fields, axis=0)
            else:
                radar_fields = radar.get_field(index, field).mean(axis=0)

            qvp.update({field:radar_fields})

    else:
        # Filtereing data based on defined gatefilter
        # If none is defined goes to else statement
        if gatefilter is not None:
            get_field = radar.get_field(index, fields)
            mask_field = np.ma.masked_where(gatefilter.gate_excluded[radar_slice],
                                            get_field)
            radar_field = np.ma.mean(mask_field, axis=0)
        else:
            radar_field = radar.get_field(index, fields).mean(axis=0)

        qvp.update({fields:radar_field})

    # Adding range, time, and height fields
    qvp.update({'range': radar.range['data'], 'time': radar.time})
    _, _, z = antenna_to_cartesian(qvp['range']/1000.0, 0.0,
                                   radar.fixed_angle['data'][index])
    qvp.update({'height': z})
    return qvp



"""
Routines for reading sinarame_H5 files.

"""

from datetime import datetime
import glob
import os

try:
    from netcdftime import utime
except ImportError:
    from cftime import utime

import numpy as np

try:
    import h5py
    _H5PY_AVAILABLE = True
except ImportError:
    _H5PY_AVAILABLE = False

from ..config import FileMetadata, get_fillvalue
from ..io.common import make_time_unit_str, _test_arguments
from ..io import write_cfradial
from ..core.radar import Radar
from ..exceptions import MissingOptionalDependency


SINARAME_H5_FIELD_NAMES = {
    'TH': 'total_power',        # uncorrected reflectivity, horizontal
    'TV': 'total_power_vertical',        # uncorrected reflectivity, vertical
    'DBZH': 'reflectivity',     # corrected reflectivity, horizontal
    'DBZV': 'reflectivity',     # corrected reflectivity, vertical
    'CM': 'clutter mask',
    'ZDR': 'differential_reflectivity',     # differential reflectivity
    'RHOHV': 'cross_correlation_ratio',
    'LDR': 'linear_polarization_ratio',
    'PHIDP': 'differential_phase',
    'KDP': 'specific_differential_phase',
    'SQI': 'normalized_coherent_power',
    'SNR': 'signal_to_noise_ratio',
    'VRAD': 'velocity',
    'WRAD': 'spectrum_width',
    'QIND': 'quality_index',
}


def read_sinarame_h5(filename, field_names=None, additional_metadata=None,
                     file_field_names=False, exclude_fields=None,
                     include_fields=None, **kwargs):
    """
    Read a SINARAME_H5 file.

    Parameters
    ----------
    filename : str
        Name of the SINARAME_H5 file to read.
    field_names : dict, optional
        Dictionary mapping SINARAME_H5 field names to radar field names. If a
        data type found in the file does not appear in this dictionary or has
        a value of None it will not be placed in the radar.fields dictionary.
        A value of None, the default, will use the mapping defined in the
        Py-ART configuration file.
    additional_metadata : dict of dicts, optional
        Dictionary of dictionaries to retrieve metadata from during this read.
        This metadata is not used during any successive file reads unless
        explicitly included.  A value of None, the default, will not
        introduct any addition metadata and the file specific or default
        metadata as specified by the Py-ART configuration file will be used.
    file_field_names : bool, optional
        True to use the MDV data type names for the field names. If this
        case the field_names parameter is ignored. The field dictionary will
        likely only have a 'data' key, unless the fields are defined in
        `additional_metadata`.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields specified by include_fields.
    include_fields : list or None, optional
        List of fields to include from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields not specified by exclude_fields.


    Returns
    -------
    radar : Radar
        Radar object containing data from SINARAME_H5 file.

    """
    # TODO before moving to pyart.io
    # * unit test
    # * add default field mapping, etc to default config
    # * auto-detect file type with pyart.io.read function
    # * instrument parameters
    # * add additional checks for HOW attributes
    # * support for other objects (SCAN, XSEC)

    # check that h5py is available
    if not _H5PY_AVAILABLE:
        raise MissingOptionalDependency(
            "h5py is required to use read_sinarame_h5 but is not installed")

    # test for non empty kwargs
    _test_arguments(kwargs)

    # create metadata retrieval object
    if field_names is None:
        field_names = SINARAME_H5_FIELD_NAMES
    filemetadata = FileMetadata('SINARAME_h5', field_names,
                                additional_metadata,
                                file_field_names, exclude_fields,
                                include_fields)

    # open the file
    hfile = h5py.File(filename, 'r')
    SINARAME_object = _to_str(hfile['what'].attrs['object'])
    if SINARAME_object not in ['PVOL', 'SCAN', 'ELEV', 'AZIM']:
        raise NotImplementedError(
            'object: %s not implemented.' % (SINARAME_object))

    # determine the number of sweeps by the number of groups which
    # begin with dataset
    datasets = [k for k in hfile if k.startswith('dataset')]
    datasets.sort(key=lambda x: int(x[7:]))
    nsweeps = len(datasets)

    # latitude, longitude and altitude
    latitude = filemetadata('latitude')
    longitude = filemetadata('longitude')
    altitude = filemetadata('altitude')

    h_where = hfile['where'].attrs
    latitude['data'] = np.array([h_where['lat']], dtype='float64')
    longitude['data'] = np.array([h_where['lon']], dtype='float64')
    altitude['data'] = np.array([h_where['height']], dtype='float64')

    # metadata
    metadata = filemetadata('metadata')
    metadata['source'] = _to_str(hfile['what'].attrs['source'])
    metadata['original_container'] = 'SINARAME_h5'
    metadata['SINARAME_conventions'] = _to_str(hfile.attrs['Conventions'])

    h_what = hfile['what'].attrs
    metadata['version'] = _to_str(h_what['version'])
    metadata['source'] = _to_str(h_what['source'])

    try:
        ds1_how = hfile[datasets[0]]['how'].attrs
    except KeyError:
        # if no how group exists mock it with an empty dictionary
        ds1_how = {}
    if 'system' in ds1_how:
        metadata['system'] = ds1_how['system']
    if 'software' in ds1_how:
        metadata['software'] = ds1_how['software']
    if 'sw_version' in ds1_how:
        metadata['sw_version'] = ds1_how['sw_version']

    # sweep_start_ray_index, sweep_end_ray_index
    sweep_start_ray_index = filemetadata('sweep_start_ray_index')
    sweep_end_ray_index = filemetadata('sweep_end_ray_index')

    if SINARAME_object in ['AZIM', 'SCAN', 'PVOL']:
        rays_per_sweep = [
            int(hfile[d]['where'].attrs['nrays']) for d in datasets]
    elif SINARAME_object == 'ELEV':
        rays_per_sweep = [
            int(hfile[d]['where'].attrs['angles'].size) for d in datasets]
    total_rays = sum(rays_per_sweep)
    ssri = np.cumsum(np.append([0], rays_per_sweep[:-1])).astype('int32')
    seri = np.cumsum(rays_per_sweep).astype('int32') - 1
    sweep_start_ray_index['data'] = ssri
    sweep_end_ray_index['data'] = seri

    # sweep_number
    sweep_number = filemetadata('sweep_number')
    sweep_number['data'] = np.arange(nsweeps, dtype='int32')

    # sweep_mode
    sweep_mode = filemetadata('sweep_mode')
    sweep_mode['data'] = np.array(nsweeps * ['azimuth_surveillance'])

    # scan_type
    if SINARAME_object == 'ELEV':
        scan_type = 'rhi'
    else:
        scan_type = 'ppi'

    # fixed_angle
    fixed_angle = filemetadata('fixed_angle')
    if SINARAME_object == 'ELEV':
        sweep_el = [hfile[d]['where'].attrs['az_angle'] for d in datasets]
    else:
        sweep_el = [hfile[d]['where'].attrs['elangle'] for d in datasets]
    fixed_angle['data'] = np.array(sweep_el, dtype='float32')

    # elevation
    elevation = filemetadata('elevation')
    if 'elangles' in ds1_how:
        edata = np.empty(total_rays, dtype='float32')
        for d, start, stop in zip(datasets, ssri, seri):
            edata[start:stop+1] = hfile[d]['how'].attrs['elangles'][:]
        elevation['data'] = edata
    elif SINARAME_object == 'ELEV':
        edata = np.empty(total_rays, dtype='float32')
        for d, start, stop in zip(datasets, ssri, seri):
            edata[start:stop+1] = hfile[d]['where'].attrs['angles'][:]
        elevation['data'] = edata
    else:
        elevation['data'] = np.repeat(sweep_el, rays_per_sweep)

    # range
    _range = filemetadata('range')
    if 'rstart' in hfile['dataset1/where'].attrs:
        # derive range from rstart and rscale attributes if available

        # check that the gate spacing is constant between sweeps
        rstart = [hfile[d]['where'].attrs['rstart'] for d in datasets]
        if any(rstart != rstart[0]):
            raise ValueError('range start changes between sweeps')
        rscale = [hfile[d]['where'].attrs['rscale'] for d in datasets]
        if any(rscale != rscale[0]):
            raise ValueError('range scale changes between sweeps')
        nbins = int(hfile['dataset1']['where'].attrs['nbins'])
        _range['data'] = (np.arange(nbins, dtype='float32') * rscale[0] +
                          rstart[0])
        _range['meters_to_center_of_first_gate'] = rstart[0]
        _range['meters_between_gates'] = float(rscale[0])
    else:
        # if not defined use range attribute which defines the maximum range in
        # km.  There is no information on the starting location of the range
        # bins so we assume this to be 0.
        # This most often occurs in RHI files, which technically do not meet
        # the ODIM 2.2 specs. Section 7.4 requires that these files include
        # the where/rstart, where/rscale and where/nbins attributes.
        max_range = [hfile[d]['where'].attrs['range'] for d in datasets]
        if any(max_range != max_range[0]):
            raise ValueError('maximum range changes between sweeps')
        # nbins is required
        nbins = hfile['dataset1/data1/data'].shape[1]
        _range['data'] = np.linspace(
            0, max_range[0] * 1000., nbins, dtype='float32')
        _range['meters_to_center_of_first_gate'] = 0
        _range['meters_between_gates'] = max_range[0] * 1000. / nbins

    # azimuth
    azimuth = filemetadata('azimuth')
    az_data = np.ones((total_rays, ), dtype='float32')
    for dset, start, stop in zip(datasets, ssri, seri):
        if SINARAME_object == 'ELEV':
            # all azimuth angles are the sweep azimuth angle
            sweep_az = hfile[dset]['where'].attrs['az_angle']
        elif SINARAME_object == 'AZIM':
            # Sector azimuths are specified in the startaz and stopaz
            # attribute of dataset/where.
            # Assume that the azimuth angles do not pass through 0/360 degrees
            startaz = hfile[dset]['where'].attrs['startaz']
            stopaz = hfile[dset]['where'].attrs['stopaz']
            nrays = stop - start + 1
            sweep_az = np.linspace(startaz, stopaz, nrays, endpoint=True)
        elif ('startazA' in ds1_how) and ('stopazA' in ds1_how):
            # average between start and stop azimuth angles
            startaz = hfile[dset]['how'].attrs['startazA']
            stopaz = hfile[dset]['how'].attrs['stopazA']
            sweep_az = np.angle(
                (np.exp(1.j*np.deg2rad(startaz)) +
                 np.exp(1.j*np.deg2rad(stopaz))) / 2., deg=True)
        else:
            # according to section 5.1 the first ray points north (0 degrees)
            # and proceeds clockwise for a complete 360 rotation.
            nrays = stop - start + 1
            sweep_az = np.linspace(0, 360, nrays, endpoint=False)
        az_data[start:stop+1] = sweep_az
    azimuth['data'] = az_data

    # time
    _time = filemetadata('time')
    if ('startazT' in ds1_how) and ('stopazT' in ds1_how):
        # average between startazT and stopazT
        t_data = np.empty((total_rays, ), dtype='float32')
        for dset, start, stop in zip(datasets, ssri, seri):
            t_start = hfile[dset]['how'].attrs['startazT']
            t_stop = hfile[dset]['how'].attrs['stopazT']
            t_data[start:stop+1] = (t_start + t_stop) / 2
        start_epoch = t_data.min()
        start_time = datetime.utcfromtimestamp(start_epoch)
        _time['units'] = make_time_unit_str(start_time)
        _time['data'] = t_data - start_epoch
    else:
        t_data = np.empty((total_rays, ), dtype='int32')
        # interpolate between each sweep starting and ending time
        for dset, start, stop in zip(datasets, ssri, seri):
            dset_what = hfile[dset]['what'].attrs
            start_str = _to_str(
                dset_what['startdate'] + dset_what['starttime'])
            end_str = _to_str(dset_what['enddate'] + dset_what['endtime'])
            start_dt = datetime.strptime(start_str, '%Y%m%d%H%M%S')
            end_dt = datetime.strptime(end_str, '%Y%m%d%H%M%S')

            time_delta = end_dt - start_dt
            delta_seconds = time_delta.seconds + time_delta.days * 3600 * 24
            rays = stop - start + 1
            sweep_start_epoch = (
                start_dt - datetime(1970, 1, 1)).total_seconds()
            t_data[start:stop+1] = (sweep_start_epoch +
                                    np.linspace(0, delta_seconds, rays))
        start_epoch = t_data.min()
        start_time = datetime.utcfromtimestamp(start_epoch)
        _time['units'] = make_time_unit_str(start_time)
        _time['data'] = (t_data - start_epoch).astype('float32')

    # fields
    fields = {}
    h_field_keys = [k for k in hfile['dataset1'] if k.startswith('data')]
    SINARAME_fields = [hfile['dataset1'][d]['what'].attrs['quantity'] for d in
                       h_field_keys]
    for SINARAME_field, h_field_key in zip(SINARAME_fields, h_field_keys):
        field_name = filemetadata.get_field_name(_to_str(SINARAME_field))
        if field_name is None:
            continue
        fdata = np.ma.zeros((total_rays, nbins), dtype='float32')
        start = 0
        # loop over the sweeps, copy data into correct location in data array
        for dset, rays_in_sweep in zip(datasets, rays_per_sweep):
            sweep_data = _get_SINARAME_h5_sweep_data(hfile[dset][h_field_key])
            sweep_nbins = sweep_data.shape[1]
            fdata[start:start + rays_in_sweep, :sweep_nbins] = sweep_data[:]
            start += rays_in_sweep
        # create field dictionary
        field_dic = filemetadata(field_name)
        field_dic['data'] = fdata
        field_dic['_FillValue'] = get_fillvalue()
        fields[field_name] = field_dic
        # Add missing metadata
        if file_field_names:
            fields[field_name].update(
                filemetadata.get_metadata(field_names[field_name]))

    # instrument_parameters
    instrument_parameters = None

    return Radar(
        _time, _range, fields, metadata, scan_type,
        latitude, longitude, altitude,
        sweep_number, sweep_mode, fixed_angle, sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth, elevation,
        instrument_parameters=instrument_parameters)


def write_sinarame_cfradial(path):
    """
    This function takes SINARAME_H5 files (where every file has only one field
    and one volume) from a folder and writes a CfRadial file for each volume
    including all fields.

    Parameters
    ----------
    path : str
        Where the SINARAME_H5 files are.

    """

    path_user = os.path.expanduser(path)

    TH_list = glob.glob(path_user + '/*_TH_*.H5')

    file_date = [i.split('_')[-1][9:-4] for i in TH_list]
    file_date.sort()

    for i in file_date:
        files = glob.glob(path_user + '/*' + i + 'Z.H5')

        # I want it to start with TV or TH because of some range issue
        files.sort(reverse=True)
        files.sort(key=lambda x: len(x.split('_')[-2]))

        for j in np.arange(len(files)):
            basename = os.path.basename(files[j])
            bs = basename.split('_')
            base1 = '{b1}_{b2}_{b3}_{fn}_{b4}'.format(
                b1=bs[0], b2=bs[1], b3=bs[2], fn=bs[3], b4=bs[4])
            file = '{path}/{base1}'.format(path=path_user, base1=base1)

            if j == 0:
                try:
                    radar = read_sinarame_h5(file, file_field_names=True)
                    azi_shape, range_shape = radar.fields['TV']['data'].shape

                except ValueError:
                    print("x - this Radar wasn't created",
                          base1, sep='\t')

                # In case the H5 file was badly created with bbufr it
                # throws a KeyError
                except KeyError:
                    print("x - Wrong BUFR conversion", base1, sep='\t')

            else:
                try:
                    radar_tmp = read_sinarame_h5(file, file_field_names=True)

                    field = radar_tmp.fields.keys()[0]

                    # Add missing gates
                    if radar_tmp.fields[field]['data'].shape[1] != range_shape:
                        n_missing_gates = \
                            (range_shape
                             - radar_tmp.fields[field]['data'].shape[1])
                        fill = np.ma.masked_all((azi_shape, n_missing_gates))
                        data = np.ma.concatenate(
                            [radar_tmp.fields[field]['data'], fill], 1)
                        radar_tmp.fields[field]['data'] = data

                    radar.fields.update(radar_tmp.fields)

                # Same as above, bbufr conversion errors
                except KeyError:
                    print("x - Wrong BUFR conversion", base1, sep='\t')

                except ValueError:
                    print("x - this Radar wasn't created",
                          base1, sep='\t')

                # In case the first file didn't create a Radar object
                except NameError:
                    print("Radar didn't exist, creating")
                    radar = read_sinarame_h5(file, file_field_names=True)

        cal_temps = u"gregorian"
        cdftime = utime(radar.time['units'])

        time1 = cdftime.num2date(radar.time['data'][0]).strftime(
            '%Y%m%d_%H%M%S')
        time2 = cdftime.num2date(radar.time['data'][-1]).strftime(
            '%Y%m%d_%H%M%S')

        radar._DeflateLevel = 5

        # cfrad.TIME1_to_TIME2_NAME_VCP_RANGE.nc
        cffile = 'cfrad.{time1}.0000_to_{time2}.0000'\
                 '_{b1}_{est}_{ran}'.format(time1=time1, time2=time2,
                                            b1=bs[0], est=bs[1], ran=bs[2])

        print('Writing to {path}{filename}.nc'.format(path=path_user,
                                                      filename=cffile))
        write_cfradial(path_user + '/' + cffile + '.nc',
                       radar, format='NETCDF4_CLASSIC')


def _to_str(text):
    """ Convert bytes to str if necessary. """
    if hasattr(text, 'decode'):
        return text.decode('utf-8')
    else:
        return text


def _get_SINARAME_h5_sweep_data(group):
    """ Get SINARAME_H5 sweet data from an HDF5 group. """

    # mask raw data
    what = group['what']
    raw_data = group['data'][:]

    if 'nodata' in what.attrs:
        nodata = what.attrs.get('nodata')
        data = np.ma.masked_equal(raw_data, nodata)
    else:
        data = np.ma.masked_array(raw_data)
    if 'undetect' in what.attrs:
        undetect = what.attrs.get('undetect')
        data[data == undetect] = np.ma.masked

    offset = 0.0
    gain = 1.0
    if 'offset' in what.attrs:
        offset = what.attrs.get('offset')
    if 'gain' in what.attrs:
        gain = what.attrs.get('gain')
    return data * gain + offset



import os
from os.path import join
import warnings
import numpy


def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration
    from numpy.distutils.system_info import get_info, BlasNotFoundError
    config = Configuration('map', parent_package, top_path)
    config.add_data_dir('tests')
    libraries = []
    if os.name == 'posix':
        libraries.append('m')

    config.add_extension('ckdtree',
                         sources=['ckdtree.c'],
                         include_dirs=[numpy.get_include()],
                         libraries=libraries)

    config.add_extension('_load_nn_field_data',
                         sources=['_load_nn_field_data.c'],
                         include_dirs=[numpy.get_include()])

    config.add_extension('_gate_to_grid_map',
                         sources=['_gate_to_grid_map.c'],
                         libraries=libraries)

    return config

if __name__ == '__main__':
    from numpy.distutils.core import setup
    setup(**configuration(top_path='').todict())



"""
Utilities for reading CF/Radial files.

"""

import datetime
import getpass
import platform
import warnings

import netCDF4
import numpy as np

from ..config import FileMetadata
from .common import stringarray_to_chararray, _test_arguments
from ..core.radar import Radar
from ..lazydict import LazyLoadDict


# Variables and dimensions in the instrument_parameter convention and
# radar_parameters sub-convention that will be read from and written to
# CfRadial files using Py-ART.
# The meta_group attribute cannot be used to identify these parameters as
# it is often set incorrectly.
_INSTRUMENT_PARAMS_DIMS = {
    # instrument_parameters sub-convention
    'frequency': ('frequency'),
    'follow_mode': ('sweep', 'string_length'),
    'pulse_width': ('time', ),
    'prt_mode': ('sweep', 'string_length'),
    'prt': ('time', ),
    'prt_ratio': ('time', ),
    'polarization_mode': ('sweep', 'string_length'),
    'nyquist_velocity': ('time', ),
    'unambiguous_range': ('time', ),
    'n_samples': ('time', ),
    'sampling_ratio': ('time', ),
    # radar_parameters sub-convention
    'radar_antenna_gain_h': (),
    'radar_antenna_gain_v': (),
    'radar_beam_width_h': (),
    'radar_beam_width_v': (),
    'radar_receiver_bandwidth': (),
    'radar_measured_transmit_power_h': ('time', ),
    'radar_measured_transmit_power_v': ('time', ),
    'radar_rx_bandwidth': (),           # non-standard
    'measured_transmit_power_v': ('time', ),    # non-standard
    'measured_transmit_power_h': ('time', ),    # non-standard
}


def read_cfradial(filename, field_names=None, additional_metadata=None,
                  file_field_names=False, exclude_fields=None,
                  include_fields=None, delay_field_loading=False, **kwargs):
    """
    Read a Cfradial netCDF file.

    Parameters
    ----------
    filename : str
        Name of CF/Radial netCDF file to read data from.
    field_names : dict, optional
        Dictionary mapping field names in the file names to radar field names.
        Unlike other read functions, fields not in this dictionary or having a
        value of None are still included in the radar.fields dictionary, to
        exclude them use the `exclude_fields` parameter. Fields which are
        mapped by this dictionary will be renamed from key to value.
    additional_metadata : dict of dicts, optional
        This parameter is not used, it is included for uniformity.
    file_field_names : bool, optional
        True to force the use of the field names from the file in which
        case the `field_names` parameter is ignored. False will use to
        `field_names` parameter to rename fields.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields specified by include_fields.
    include_fields : list or None, optional
        List of fields to include from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields not specified by exclude_fields.
    delay_field_loading : bool
        True to delay loading of field data from the file until the 'data'
        key in a particular field dictionary is accessed. In this case
        the field attribute of the returned Radar object will contain
        LazyLoadDict objects not dict objects. Delayed field loading will not
        provide any speedup in file where the number of gates vary between
        rays (ngates_vary=True) and is not recommended.

    Returns
    -------
    radar : Radar
        Radar object.

    Notes
    -----
    This function has not been tested on "stream" Cfradial files.

    """
    # test for non empty kwargs
    _test_arguments(kwargs)

    # create metadata retrieval object
    filemetadata = FileMetadata('cfradial', field_names, additional_metadata,
                                file_field_names, exclude_fields)

    # read the data
    ncobj = netCDF4.Dataset(filename)
    ncvars = ncobj.variables

    # 4.1 Global attribute -> move to metadata dictionary
    metadata = dict([(k, getattr(ncobj, k)) for k in ncobj.ncattrs()])
    if 'n_gates_vary' in metadata:
        metadata['n_gates_vary'] = 'false'  # corrected below

    # 4.2 Dimensions (do nothing)

    # 4.3 Global variable -> move to metadata dictionary
    if 'volume_number' in ncvars:
        metadata['volume_number'] = int(ncvars['volume_number'][:])
    else:
        metadata['volume_number'] = 0

    global_vars = {'platform_type': 'fixed', 'instrument_type': 'radar',
                   'primary_axis': 'axis_z'}
    # ignore time_* global variables, these are calculated from the time
    # variable when the file is written.
    for var, default_value in global_vars.items():
        if var in ncvars:
            metadata[var] = str(netCDF4.chartostring(ncvars[var][:]))
        else:
            metadata[var] = default_value

    # 4.4 coordinate variables -> create attribute dictionaries
    time = _ncvar_to_dict(ncvars['time'])
    _range = _ncvar_to_dict(ncvars['range'])

    # 4.5 Ray dimension variables

    # 4.6 Location variables -> create attribute dictionaries
    latitude = _ncvar_to_dict(ncvars['latitude'])
    longitude = _ncvar_to_dict(ncvars['longitude'])
    altitude = _ncvar_to_dict(ncvars['altitude'])
    if 'altitude_agl' in ncvars:
        altitude_agl = _ncvar_to_dict(ncvars['altitude_agl'])
    else:
        altitude_agl = None

    # 4.7 Sweep variables -> create atrribute dictionaries
    sweep_mode = _ncvar_to_dict(ncvars['sweep_mode'])
    fixed_angle = _ncvar_to_dict(ncvars['fixed_angle'])
    sweep_start_ray_index = _ncvar_to_dict(ncvars['sweep_start_ray_index'])
    sweep_end_ray_index = _ncvar_to_dict(ncvars['sweep_end_ray_index'])

    if 'sweep_number' in ncvars:
        sweep_number = _ncvar_to_dict(ncvars['sweep_number'])
    else:
        nsweeps = len(sweep_start_ray_index['data'])
        sweep_number = filemetadata('sweep_number')
        sweep_number['data'] = np.arange(nsweeps, dtype='float32')
        warnings.warn("Warning: File violates CF/Radial convention. " +
                      "Missing sweep_number variable")

    if 'target_scan_rate' in ncvars:
        target_scan_rate = _ncvar_to_dict(ncvars['target_scan_rate'])
    else:
        target_scan_rate = None
    if 'rays_are_indexed' in ncvars:
        rays_are_indexed = _ncvar_to_dict(ncvars['rays_are_indexed'])
    else:
        rays_are_indexed = None
    if 'ray_angle_res' in ncvars:
        ray_angle_res = _ncvar_to_dict(ncvars['ray_angle_res'])
    else:
        ray_angle_res = None

    # first sweep mode determines scan_type
    try:
        mode = netCDF4.chartostring(sweep_mode['data'][0])[()].decode('utf-8')
    except AttributeError:
        # Python 3, all strings are already unicode.
        mode = netCDF4.chartostring(sweep_mode['data'][0])[()]

    # options specified in the CF/Radial standard
    if mode == 'rhi':
        scan_type = 'rhi'
    elif mode == 'vertical_pointing':
        scan_type = 'vpt'
    elif mode == 'azimuth_surveillance':
        scan_type = 'ppi'
    elif mode == 'elevation_surveillance':
        scan_type = 'rhi'
    elif mode == 'manual_ppi':
        scan_type = 'ppi'
    elif mode == 'manual_rhi':
        scan_type = 'rhi'

    # fallback types
    elif 'sur' in mode:
        scan_type = 'ppi'
    elif 'sec' in mode:
        scan_type = 'sector'
    elif 'rhi' in mode:
        scan_type = 'rhi'
    elif 'ppi' in mode:
        scan_type = 'ppi'
    else:
        scan_type = 'other'

    # 4.8 Sensor pointing variables -> create attribute dictionaries
    azimuth = _ncvar_to_dict(ncvars['azimuth'])
    elevation = _ncvar_to_dict(ncvars['elevation'])
    if 'scan_rate' in ncvars:
        scan_rate = _ncvar_to_dict(ncvars['scan_rate'])
    else:
        scan_rate = None

    if 'antenna_transition' in ncvars:
        antenna_transition = _ncvar_to_dict(ncvars['antenna_transition'])
    else:
        antenna_transition = None

    # 4.9 Moving platform geo-reference variables
    # Aircraft specific varaibles
    if 'rotation' in ncvars:
        rotation = _ncvar_to_dict(ncvars['rotation'])
    else:
        rotation = None

    if 'tilt' in ncvars:
        tilt = _ncvar_to_dict(ncvars['tilt'])
    else:
        tilt = None

    if 'roll' in ncvars:
        roll = _ncvar_to_dict(ncvars['roll'])
    else:
        roll = None

    if 'drift' in ncvars:
        drift = _ncvar_to_dict(ncvars['drift'])
    else:
        drift = None

    if 'heading' in ncvars:
        heading = _ncvar_to_dict(ncvars['heading'])
    else:
        heading = None

    if 'pitch' in ncvars:
        pitch = _ncvar_to_dict(ncvars['pitch'])
    else:
        pitch = None

    if 'georefs_applied' in ncvars:
        georefs_applied = _ncvar_to_dict(ncvars['georefs_applied'])
    else:
        georefs_applied = None

    # 4.10 Moments field data variables -> field attribute dictionary
    if 'ray_n_gates' in ncvars:
        # all variables with dimensions of n_points are fields.
        keys = [k for k, v in ncvars.items()
                if v.dimensions == ('n_points', )]
    else:
        # all variables with dimensions of 'time', 'range' are fields
        keys = [k for k, v in ncvars.items()
                if v.dimensions == ('time', 'range')]

    fields = {}
    for key in keys:
        field_name = filemetadata.get_field_name(key)
        if field_name is None:
            if exclude_fields is not None and key in exclude_fields:
                if key not in include_fields:
                    continue
            if include_fields is None or key in include_fields:
                field_name = key
            else:
                continue
        fields[field_name] = _ncvar_to_dict(ncvars[key], delay_field_loading)

    if 'ray_n_gates' in ncvars:
        shape = (len(ncvars['time']), len(ncvars['range']))
        ray_n_gates = ncvars['ray_n_gates'][:]
        ray_start_index = ncvars['ray_start_index'][:]
        for dic in fields.values():
            _unpack_variable_gate_field_dic(
                dic, shape, ray_n_gates, ray_start_index)

    # 4.5 instrument_parameters sub-convention -> instrument_parameters dict
    # 4.6 radar_parameters sub-convention -> instrument_parameters dict
    keys = [k for k in _INSTRUMENT_PARAMS_DIMS.keys() if k in ncvars]
    instrument_parameters = dict((k, _ncvar_to_dict(ncvars[k])) for k in keys)
    if instrument_parameters == {}:  # if no parameters set to None
        instrument_parameters = None

    # 4.7 lidar_parameters sub-convention -> skip

    # 4.8 radar_calibration sub-convention -> radar_calibration
    keys = _find_all_meta_group_vars(ncvars, 'radar_calibration')
    radar_calibration = dict((k, _ncvar_to_dict(ncvars[k])) for k in keys)
    if radar_calibration == {}:
        radar_calibration = None

    # do not close file is field loading is delayed
    if not delay_field_loading:
        ncobj.close()
    return Radar(
        time, _range, fields, metadata, scan_type,
        latitude, longitude, altitude,
        sweep_number, sweep_mode, fixed_angle, sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth, elevation,
        instrument_parameters=instrument_parameters,
        radar_calibration=radar_calibration,
        altitude_agl=altitude_agl,
        scan_rate=scan_rate,
        antenna_transition=antenna_transition,
        target_scan_rate=target_scan_rate,
        rays_are_indexed=rays_are_indexed, ray_angle_res=ray_angle_res,
        rotation=rotation, tilt=tilt, roll=roll, drift=drift, heading=heading,
        pitch=pitch, georefs_applied=georefs_applied)


def _find_all_meta_group_vars(ncvars, meta_group_name):
    """
    Return a list of all variables which are in a given meta_group.
    """
    return [k for k, v in ncvars.items() if 'meta_group' in v.ncattrs() and
            v.meta_group == meta_group_name]


def _ncvar_to_dict(ncvar, lazydict=False):
    """ Convert a NetCDF Dataset variable to a dictionary. """
    # copy all attribute except for scaling parameters
    d = dict((k, getattr(ncvar, k)) for k in ncvar.ncattrs()
             if k not in ['scale_factor', 'add_offset'])
    data_extractor = _NetCDFVariableDataExtractor(ncvar)
    if lazydict:
        d = LazyLoadDict(d)
        d.set_lazy('data', data_extractor)
    else:
        d['data'] = data_extractor()
    return d


class _NetCDFVariableDataExtractor(object):
    """
    Class facilitating on demand extraction of data from a NetCDF variable.

    Parameters
    ----------
    ncvar : netCDF4.Variable
        NetCDF Variable from which data will be extracted.

    """

    def __init__(self, ncvar):
        """ initialize the object. """
        self.ncvar = ncvar

    def __call__(self):
        """ Return an array containing data from the stored variable. """
        data = self.ncvar[:]
        if data is np.ma.masked:
            # If the data is a masked scalar, MaskedConstant is returned by
            # NetCDF4 version 1.2.3+. This object does not preserve the dtype
            # and fill_value of the original NetCDF variable and causes issues
            # in Py-ART.
            # Rather we create a masked array with a single masked value
            # with the correct dtype and fill_value.
            self.ncvar.set_auto_mask(False)
            data = np.ma.array(self.ncvar[:], mask=True)
        # Use atleast_1d to force the array to be at minimum one dimensional,
        # some version of netCDF return scalar or scalar arrays for scalar
        # NetCDF variables.
        return np.atleast_1d(data)


def _unpack_variable_gate_field_dic(
        dic, shape, ray_n_gates, ray_start_index):
    """ Create a 2D array from a 1D field data, dic update in place. """
    fdata = dic['data']
    data = np.ma.masked_all(shape, dtype=fdata.dtype)
    for i, (gates, idx) in enumerate(zip(ray_n_gates, ray_start_index)):
        data[i, :gates] = fdata[idx:idx+gates]
    dic['data'] = data
    return


def write_cfradial(filename, radar, format='NETCDF4', time_reference=None,
                   arm_time_variables=False):
    """
    Write a Radar object to a CF/Radial compliant netCDF file.

    The files produced by this routine follow the `CF/Radial standard`_.
    Attempts are also made to to meet many of the standards outlined in the
    `ARM Data File Standards`_.

    .. _CF/Radial standard: http://www.ral.ucar.edu/projects/titan/docs/radial_formats/cfradial.html
    .. _ARM Data File Standards: https://docs.google.com/document/d/1gBMw4Kje6v8LBlsrjaGFfSLoU0jRx-07TIazpthZGt0/edit?pli=1

    To control how the netCDF variables are created, set any of the following
    keys in the radar attribute dictionaries.

        * _Zlib
        * _DeflateLevel
        * _Shuffle
        * _Fletcher32
        * _Continguous
        * _ChunkSizes
        * _Endianness
        * _Least_significant_digit
        * _FillValue

    See the netCDF4 documentation for details on these settings.

    Parameters
    ----------
    filename : str
        Filename to create.
    radar : Radar
        Radar object.
    format : str, optional
        NetCDF format, one of 'NETCDF4', 'NETCDF4_CLASSIC',
        'NETCDF3_CLASSIC' or 'NETCDF3_64BIT'. See netCDF4 documentation for
        details.
    time_reference : bool
        True to include a time_reference variable, False will not include
        this variable. The default, None, will include the time_reference
        variable when the first time value is non-zero.
    arm_time_variables : bool
        True to create the ARM standard time variables base_time and
        time_offset, False will not create these variables.

    """
    dataset = netCDF4.Dataset(filename, 'w', format=format)

    # determine the maximum string length
    max_str_len = len(radar.sweep_mode['data'][0])
    for k in ['follow_mode', 'prt_mode', 'polarization_mode']:
        if ((radar.instrument_parameters is not None) and
                (k in radar.instrument_parameters)):
            sdim_length = len(radar.instrument_parameters[k]['data'][0])
            max_str_len = max(max_str_len, sdim_length)
    str_len = max(max_str_len, 32)      # minimum string legth of 32

    # create time, range and sweep dimensions
    dataset.createDimension('time', None)
    dataset.createDimension('range', radar.ngates)
    dataset.createDimension('sweep', radar.nsweeps)
    dataset.createDimension('string_length', str_len)

    # global attributes
    # remove global variables from copy of metadata
    metadata_copy = dict(radar.metadata)
    global_variables = ['volume_number', 'platform_type', 'instrument_type',
                        'primary_axis', 'time_coverage_start',
                        'time_coverage_end', 'time_reference']
    for var in global_variables:
        if var in metadata_copy:
            metadata_copy.pop(var)

    # determine the history attribute if it doesn't exist, save for
    # the last attribute.
    if 'history' in metadata_copy:
        history = metadata_copy.pop('history')
    else:
        user = getpass.getuser()
        node = platform.node()
        time_str = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%SZ')
        t = (user, node, time_str)
        history = 'created by %s on %s at %s using Py-ART' % (t)

    dataset.setncatts(metadata_copy)

    if 'Conventions' not in dataset.ncattrs():
        dataset.setncattr('Conventions', "CF/Radial")

    if 'field_names' not in dataset.ncattrs():
        dataset.setncattr('field_names', ', '.join(radar.fields.keys()))

    # history should be the last attribute, ARM standard
    dataset.setncattr('history', history)

    # arm time variables base_time and time_offset if requested
    if arm_time_variables:
        dt = netCDF4.num2date(radar.time['data'][0], radar.time['units'])
        td = dt - datetime.datetime.utcfromtimestamp(0)
        base_time = {
            'data': np.array([td.seconds + td.days * 24 * 3600], 'int32'),
            'string': dt.strftime('%d-%b-%Y,%H:%M:%S GMT'),
            'units': 'seconds since 1970-1-1 0:00:00 0:00',
            'ancillary_variables': 'time_offset',
            'long_name': 'Base time in Epoch',
        }
        _create_ncvar(base_time, dataset, 'base_time', ())

        time_offset = {
            'data': radar.time['data'],
            'long_name': 'Time offset from base_time',
            'units': radar.time['units'].replace('T', ' ').replace('Z', ''),
            'ancillary_variables': 'time_offset',
            'calendar': 'gregorian',
        }
        _create_ncvar(time_offset, dataset, 'time_offset', ('time', ))

    # standard variables
    _create_ncvar(radar.time, dataset, 'time', ('time', ))
    _create_ncvar(radar.range, dataset, 'range', ('range', ))
    _create_ncvar(radar.azimuth, dataset, 'azimuth', ('time', ))
    _create_ncvar(radar.elevation, dataset, 'elevation', ('time', ))

    # optional sensor pointing variables
    if radar.scan_rate is not None:
        _create_ncvar(radar.scan_rate, dataset, 'scan_rate', ('time', ))
    if radar.antenna_transition is not None:
        _create_ncvar(radar.antenna_transition, dataset,
                      'antenna_transition', ('time', ))

    # fields
    for field, dic in radar.fields.items():
        _create_ncvar(dic, dataset, field, ('time', 'range'))

    # sweep parameters
    _create_ncvar(radar.sweep_number, dataset, 'sweep_number', ('sweep', ))
    _create_ncvar(radar.fixed_angle, dataset, 'fixed_angle', ('sweep', ))
    _create_ncvar(radar.sweep_start_ray_index, dataset,
                  'sweep_start_ray_index', ('sweep', ))
    _create_ncvar(radar.sweep_end_ray_index, dataset,
                  'sweep_end_ray_index', ('sweep', ))
    _create_ncvar(radar.sweep_mode, dataset, 'sweep_mode',
                  ('sweep', 'string_length'))
    if radar.target_scan_rate is not None:
        _create_ncvar(radar.target_scan_rate, dataset, 'target_scan_rate',
                      ('sweep', ))
    if radar.rays_are_indexed is not None:
        _create_ncvar(radar.rays_are_indexed, dataset, 'rays_are_indexed',
                      ('sweep', 'string_length'))
    if radar.ray_angle_res is not None:
        _create_ncvar(radar.ray_angle_res, dataset, 'ray_angle_res',
                      ('sweep', ))

    # instrument_parameters
    if ((radar.instrument_parameters is not None) and
            ('frequency' in radar.instrument_parameters.keys())):
        size = len(radar.instrument_parameters['frequency']['data'])
        dataset.createDimension('frequency', size)

    if radar.instrument_parameters is not None:
        for k in radar.instrument_parameters.keys():
            if k in _INSTRUMENT_PARAMS_DIMS:
                dim = _INSTRUMENT_PARAMS_DIMS[k]
                _create_ncvar(radar.instrument_parameters[k], dataset, k, dim)
            else:
                # Do not try to write instrument parameter whose dimensions are
                # not known, rather issue a warning and skip the parameter
                message = ("Unknown instrument parameter: %s, " % (k) +
                           "not written to file.")
                warnings.warn(message)

    # radar_calibration variables
    if radar.radar_calibration is not None and radar.radar_calibration != {}:
        size = [len(d['data']) for k, d in radar.radar_calibration.items()
                if k not in ['r_calib_index', 'r_calib_time']][0]
        dataset.createDimension('r_calib', size)
        for key, dic in radar.radar_calibration.items():
            if key == 'r_calib_index':
                dims = ('time', )
            elif key == 'r_calib_time':
                dims = ('r_calib', 'string_length')
            else:
                dims = ('r_calib', )
            _create_ncvar(dic, dataset, key, dims)

    # latitude, longitude, altitude, altitude_agl
    if radar.latitude['data'].size == 1:
        # stationary platform
        _create_ncvar(radar.latitude, dataset, 'latitude', ())
        _create_ncvar(radar.longitude, dataset, 'longitude', ())
        _create_ncvar(radar.altitude, dataset, 'altitude', ())
        if radar.altitude_agl is not None:
            _create_ncvar(radar.altitude_agl, dataset, 'altitude_agl', ())
    else:
        # moving platform
        _create_ncvar(radar.latitude, dataset, 'latitude', ('time', ))
        _create_ncvar(radar.longitude, dataset, 'longitude', ('time', ))
        _create_ncvar(radar.altitude, dataset, 'altitude', ('time', ))
        if radar.altitude_agl is not None:
            _create_ncvar(radar.altitude_agl, dataset, 'altitude_agl',
                          ('time', ))

    # time_coverage_start and time_coverage_end variables
    time_dim = ('string_length', )
    units = radar.time['units']
    start_dt = netCDF4.num2date(radar.time['data'][0], units)
    if start_dt.microsecond != 0:
        # truncate to nearest second
        start_dt -= datetime.timedelta(microseconds=start_dt.microsecond)
    end_dt = netCDF4.num2date(radar.time['data'][-1], units)
    if end_dt.microsecond != 0:
        # round up to next second
        end_dt += (datetime.timedelta(seconds=1) -
                   datetime.timedelta(microseconds=end_dt.microsecond))
    start_dic = {'data': np.array(start_dt.strftime('%Y-%m-%dT%H:%M:%SZ'),
                                  dtype='S'),
                 'long_name': 'UTC time of first ray in the file',
                 'units': 'unitless'}
    end_dic = {'data': np.array(end_dt.strftime('%Y-%m-%dT%H:%M:%SZ'),
                                dtype='S'),
               'long_name': 'UTC time of last ray in the file',
               'units': 'unitless'}
    _create_ncvar(start_dic, dataset, 'time_coverage_start', time_dim)
    _create_ncvar(end_dic, dataset, 'time_coverage_end', time_dim)

    # time_reference is required or requested.
    if time_reference is None:
        if radar.time['data'][0] == 0:
            time_reference = False
        else:
            time_reference = True
    if time_reference:
        ref_dic = {'data': np.array(radar.time['units'][-20:], dtype='S'),
                   'long_name': 'UTC time reference',
                   'units': 'unitless'}
        _create_ncvar(ref_dic, dataset, 'time_reference', time_dim)

    # global variables
    # volume_number, required
    vol_dic = {'long_name': 'Volume number', 'units': 'unitless'}
    if 'volume_number' in radar.metadata:
        vol_dic['data'] = np.array([radar.metadata['volume_number']],
                                   dtype='int32')
    else:
        vol_dic['data'] = np.array([0], dtype='int32')
    _create_ncvar(vol_dic, dataset, 'volume_number', ())

    # platform_type, optional
    if 'platform_type' in radar.metadata:
        dic = {'long_name': 'Platform type',
               'data': np.array(radar.metadata['platform_type'], dtype='S')}
        _create_ncvar(dic, dataset, 'platform_type', ('string_length', ))

    # instrument_type, optional
    if 'instrument_type' in radar.metadata:
        dic = {'long_name': 'Instrument type',
               'data': np.array(radar.metadata['instrument_type'], dtype='S')}
        _create_ncvar(dic, dataset, 'instrument_type', ('string_length', ))

    # primary_axis, optional
    if 'primary_axis' in radar.metadata:
        dic = {'long_name': 'Primary axis',
               'data': np.array(radar.metadata['primary_axis'], dtype='S')}
        _create_ncvar(dic, dataset, 'primary_axis', ('string_length', ))

    # moving platform geo-reference variables
    if radar.rotation is not None:
        _create_ncvar(radar.rotation, dataset, 'rotation', ('time', ))

    if radar.tilt is not None:
        _create_ncvar(radar.tilt, dataset, 'tilt', ('time', ))

    if radar.roll is not None:
        _create_ncvar(radar.roll, dataset, 'roll', ('time', ))

    if radar.drift is not None:
        _create_ncvar(radar.drift, dataset, 'drift', ('time', ))

    if radar.heading is not None:
        _create_ncvar(radar.heading, dataset, 'heading', ('time', ))

    if radar.pitch is not None:
        _create_ncvar(radar.pitch, dataset, 'pitch', ('time', ))

    if radar.georefs_applied is not None:
        _create_ncvar(radar.georefs_applied, dataset, 'georefs_applied',
                      ('time', ))

    dataset.close()


def _create_ncvar(dic, dataset, name, dimensions):
    """
    Create and fill a Variable in a netCDF Dataset object.

    Parameters
    ----------
    dic : dict
        Radar dictionary to containing variable data and meta-data.
    dataset : Dataset
        NetCDF dataset to create variable in.
    name : str
        Name of variable to create.
    dimension : tuple of str
        Dimension of variable.

    """
    # create array from list, etc.
    data = dic['data']
    if isinstance(data, np.ndarray) is not True:
        warnings.warn("Warning, converting non-array to array:%s" % name)
        data = np.array(data)

    # convert string/unicode arrays to character arrays
    if data.dtype.char == 'U':  # cast unicode arrays to char arrays
        data = data.astype('S')
    if data.dtype.char == 'S' and data.dtype != 'S1':
        data = stringarray_to_chararray(data)

    # determine netCDF variable arguments
    special_keys = {
        # dictionary keys which can be used to change the default values of
        # createVariable arguments, some of these map to netCDF special
        # attributes, other are Py-ART conventions.
        '_Zlib': 'zlib',
        '_DeflateLevel': 'complevel',
        '_Shuffle': 'shuffle',
        '_Fletcher32': 'fletcher32',
        '_Continguous': 'contiguous',
        '_ChunkSizes': 'chunksizes',
        '_Endianness': 'endian',
        '_Least_significant_digit': 'least_significant_digit',
        '_FillValue': 'fill_value',
    }
    kwargs = {'zlib': True}  # default is to use compression
    for dic_key, kwargs_key in special_keys.items():
        if dic_key in dic:
            kwargs[kwargs_key] = dic[dic_key]

    # the _Write_as_dtype key can be used to specify the netCDF dtype
    if '_Write_as_dtype' in dic:
        dtype = np.dtype(dic['_Write_as_dtype'])
        if np.issubdtype(dtype, np.integer):
            if 'scale_factor' not in dic and 'add_offset' not in dic:
                # calculate scale and offset
                scale, offset, fill = _calculate_scale_and_offset(dic, dtype)
                dic['scale_factor'] = scale
                dic['add_offset'] = offset
                dic['_FillValue'] = fill
                kwargs['fill_value'] = fill
    else:
        dtype = data.dtype

    # create the dataset variable
    ncvar = dataset.createVariable(name, dtype, dimensions, **kwargs)

    # long_name attribute first if present, ARM standard
    if 'long_name' in dic.keys():
        ncvar.setncattr('long_name', dic['long_name'])

    # units attribute second if present, ARM standard
    if 'units' in dic.keys():
        ncvar.setncattr('units', dic['units'])

    # set all attributes
    for key, value in dic.items():
        if key in special_keys.keys():
            continue
        if key in ['data', 'long_name', 'units']:
            continue
        ncvar.setncattr(key, value)

    # set the data
    if data.shape == ():
        data.shape = (1,)
    if data.dtype == 'S1':  # string/char arrays
        # KLUDGE netCDF4 version around 1.1.6 do not expand an ellipsis
        # to zero dimensions (Issue #371 of netcdf4-python).
        # Solution is so we treat 1 dimensional string dimensions explicitly.
        if ncvar.ndim == 1:
            ncvar[:data.shape[-1]] = data[:]
        else:
            ncvar[..., :data.shape[-1]] = data[:]
    else:
        ncvar[:] = data[:]


def _calculate_scale_and_offset(dic, dtype, minimum=None, maximum=None):
    """
    Calculate appropriated 'scale_factor' and 'add_offset' for nc variable in
    dic in order to scaling to fit dtype range.

    Parameters
    ----------
    dic : dict
        Radar dictionary containing variable data and meta-data.
    dtype : Numpy Dtype
        Integer numpy dtype to map to.
    minimum, maximum : float
        Greatest and smallest values in the data, those values will be mapped
        to the smallest+1 and greates values that dtype can hold.
        If equal to None, numpy.amin and numpy.amax will be used on the data
        contained in dic to determine these values.

    """
    if "_FillValue" in dic:
        fillvalue = dic["_FillValue"]
    else:
        fillvalue = np.NaN

    data = dic['data'].copy()
    data = np.ma.array(data, mask=(~np.isfinite(data) | (data == fillvalue)))

    if minimum is None:
        minimum = np.amin(data)
    if maximum is None:
        maximum = np.amax(data)

    if maximum < minimum:
        raise ValueError(
            'Error calculating variable scaling: '
            'maximum: %f is smaller than minimum: %f' % (maximum, minimum))
    elif maximum == minimum:
        warnings.warn(
            'While calculating variable scaling: '
            'maximum: %f is equal to minimum: %f' % (maximum, minimum))
        maximum = minimum + 1

    # get max and min scaled,
    maxi = np.iinfo(dtype).max
    mini = np.iinfo(dtype).min + 1  # +1 since min will serve as the fillvalue
    scale = float(maximum - minimum) / float(maxi - mini)
    offset = minimum - mini * scale

    return scale, offset, np.iinfo(dtype).min




# THIS FILE IS GENERATED FROM PYART SETUP.PY
short_version = '1.11.1'
version = '1.11.1'
full_version = '1.11.1'
git_revision = 'd216a2750da3014b3f387f3915b70730c64763cb'
release = True

if not release:
    version = full_version



"""
Creating plots of Radar and Grid fields.

There are also Radar related colormaps and colorblind friendly radar
colormaps for plotting.

Available colormaps, reversed versions (_r) are also provided, these
colormaps are available within matplotlib with names 'pyart_COLORMAP':

    * BlueBrown10
    * BlueBrown11
    * BrBu10
    * BrBu12
    * Bu10
    * Bu7
    * BuDOr12
    * BuDOr18
    * BuDRd12
    * BuDRd18
    * BuGr14
    * BuGy8
    * BuOr10
    * BuOr12
    * BuOr8
    * BuOrR14
    * Carbone11
    * Carbone17
    * Carbone42
    * Cat12
    * EWilson17
    * GrMg16
    * Gray5
    * Gray9
    * NWSRef
    * NWSVel
    * NWS_SPW
    * PD17
    * RRate11
    * RdYlBu11b
    * RefDiff
    * SCook18
    * StepSeq25
    * SymGray12
    * Theodore16
    * Wild25

Colorblind friendly

    * LangRainbow12
    * HomeyerRainbow
    * balance

"""

from .radardisplay import RadarDisplay
from . import cm
from . import cm_colorblind
from .radardisplay_airborne import AirborneRadarDisplay
from .gridmapdisplay import GridMapDisplay
from .gridmapdisplay_basemap import GridMapDisplayBasemap
from .radarmapdisplay import RadarMapDisplay
from .radarmapdisplay_basemap import RadarMapDisplayBasemap

__all__ = [s for s in dir() if not s.startswith('_')]



"""
Functions that retrieve height of the gates and the profile interpolated
onto the radar gates.

"""

try:
    from netCDF4 import num2date, datetime, DatetimeGregorian
except ImportError:
    from cftime import num2date, datetime, DatetimeGregorian

import numpy as np
from scipy import interpolate

from ..config import get_fillvalue, get_metadata, get_field_name
from ..core.transforms import antenna_to_cartesian


def map_profile_to_gates(profile, heights, radar, toa=None,
                         profile_field=None, height_field=None):
    """
    Given a profile of a variable map it to the gates of radar assuming 4/3Re.

    Parameters
    ----------
    profile : array
        Profile array to map.
    heights : array
        Monotonically increasing heights in meters with same shape as profile.
    radar : Radar
        Radar to map to.
    toa : float, optional
        Top of atmosphere, where to use profile up to. If None check for
        mask and use lowest element, if no mask uses whole profile.
    height_field : str, optional
        Name to use for height field metadata. None will use the default field
        name from the Py-ART configuration file.
    profile_field : str, optional
        Name to use for interpolate profile field metadata. None will use the
        default field name from the Py-ART configuration file.

    Returns
    -------
    height_dict, profile_dict : dict
        Field dictionaries containing the height of the gates and the profile
        interpolated onto the radar gates.

    """
    # retrieve the Z coordinates of the radar gates
    rg, azg = np.meshgrid(radar.range['data'], radar.azimuth['data'])
    rg, eleg = np.meshgrid(radar.range['data'], radar.elevation['data'])
    _, _, z = antenna_to_cartesian(rg / 1000.0, azg, eleg)

    # Check that z is not a MaskedArray
    if isinstance(z, np.ma.MaskedArray):
        z = z.filled(np.NaN)

    # find toa is not provided
    if toa is None:
        ismasked = np.where(np.ma.getmaskarray(profile))[0]
        if len(ismasked) == 0:
            toa = None
        else:
            toa = ismasked.min()

    # interpolate
    f_interp = interpolate.interp1d(
        heights[:toa], profile[:toa], bounds_error=False,
        fill_value=get_fillvalue())
    fld = np.ma.masked_equal(f_interp(z + radar.altitude['data'][0]),
                             get_fillvalue())

    # populate field dictionaries
    if height_field is None:
        height_field = get_field_name('height')
    height_dict = get_metadata(height_field)
    height_dict['data'] = z + radar.altitude['data'][0]

    if profile_field is None:
        profile_field = get_field_name('interpolated_profile')
    profile_dict = get_metadata(profile_field)
    profile_dict['data'] = fld

    return height_dict, profile_dict


def fetch_radar_time_profile(sonde_dset, radar, time_key='time',
                             height_key='height', nvars=None):
    """
    Extract the correct profile from a interpolated sonde.

    This is an ARM specific method which extract the correct profile out of
    netCDF Variables from a Interpolated Sonde VAP for the volume start time
    of a radar object.

    Parameters
    ----------
    sonde_dset : Dataset
        Interpolate sonde Dataset.
    radar : Radar
        Radar object from which the nearest profile will be found.
    time_key : string, optional
        Key to find a CF startard time variable.
    height_key : string, optional
        Key to find profile height data.
    nvars : list, optional
        NetCDF variable to generated profiles for. If None (the default) all
        variables with dimension of time, height will be found in ncvars.

    Returns
    -------
    return_dic : dict
        Profiles at the start time of the radar.

    """
    ncvars = sonde_dset.variables
    if nvars is None:
        time_height_shape = (len(ncvars[time_key]), len(ncvars[height_key]))
        nvars = [k for k, v in ncvars.items() if v.shape == time_height_shape]

    radar_start = num2date(radar.time['data'][0], radar.time['units'])
    radar_day_start = DatetimeGregorian(radar_start.year, radar_start.month,
                                        radar_start.day)
    seconds_since_start_of_day = (radar_start - radar_day_start).seconds
    time_index = abs(ncvars[time_key][:] - seconds_since_start_of_day).argmin()

    return_dic = dict([(key, ncvars[key][time_index, :]) for key in nvars])
    return_dic[height_key] = ncvars[height_key][:]
    return return_dic
